from __future__ import annotations

from dataclasses import dataclass, field
from typing import Sequence, TYPE_CHECKING

from ResSimpy.Nexus.DataModels.Network.NexusProc import NexusProc

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import NexusNetwork


@dataclass(kw_only=True)
class NexusProcs:
    """Class for handling procedures in the Nexus Network.

    This class is used to store and collect information about certain procedures written in the Nexus surface file.
    """
    # need field function to make a new unique empty list
    __procs: list[NexusProc] = field(default_factory=list)

    def __init__(self, parent_network: NexusNetwork) -> None:
        """Initialises the NexusProcs class.

        Args:
            parent_network (NexusNetwork): The network that the procedures are a part of.
        """
        self.__parent_network: NexusNetwork = parent_network
        self.__procs = []

    def get_all(self) -> Sequence[NexusProc]:
        """Returns a list of procs loaded from the simulator."""
        self.__parent_network.get_load_status()
        return self.__procs

    def _add_to_memory(self, procs_to_add: list[NexusProc]) -> None:
        """Adds the list of Nexus procedure objects to memory."""
        self.__procs.extend(procs_to_add)



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Sequence, TYPE_CHECKING, Optional

from ResSimpy.Nexus.DataModels.Network.NexusActivationChange import NexusActivationChange
from ResSimpy.Nexus.NexusEnums.ActivationChangeEnum import ActivationChangeEnum
from ResSimpy.Time.ISODateTime import ISODateTime

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import NexusNetwork


@dataclass(kw_only=True)
class NexusActivationChanges:
    """A class representing a list of nexus action objects."""
    __activationChanges: list[NexusActivationChange] = field(default_factory=list)

    def __init__(self, parent_network: NexusNetwork) -> None:
        """Initialises the NexusActivationChanges class.

        Args:
            parent_network (NexusNetwork): The network that the actions are a part of.
        """
        self.__parent_network: NexusNetwork = parent_network
        self.__activationChanges = []

    def get_all(self) -> Sequence[NexusActivationChange]:
        """Ensures the surface network file has been loaded."""
        self.__parent_network.get_load_status()
        return self.__activationChanges

    def _add_to_memory(self, activations_to_add: Optional[list[NexusActivationChange]]) -> None:
        """Adds the list of NexusActivationChange objects to memory."""
        if activations_to_add is None:
            return

        self.__activationChanges.extend(activations_to_add)

    def to_string_for_date(self, date: ISODateTime) -> str:
        """Outputs the activation changes for a specific date."""
        output_str = ''
        activations_for_date = [x for x in self.__activationChanges if x.iso_date == date and
                                x.change == ActivationChangeEnum.ACTIVATE]
        deactivations_for_date = [x for x in self.__activationChanges if x.iso_date == date and
                                  x.change == ActivationChangeEnum.DEACTIVATE]

        if activations_for_date:
            output_str += "ACTIVATE\nCONNECTION\n"
        for activation in activations_for_date:
            output_str += f"{activation.name}\n"
        if activations_for_date:
            output_str += "ENDACTIVATE\n"

        if deactivations_for_date:
            output_str += "DEACTIVATE\nCONNECTION\n"
        for deactivation in deactivations_for_date:
            output_str += f"{deactivation.name}\n"
        if deactivations_for_date:
            output_str += "ENDDEACTIVATE\n"

        return output_str



from __future__ import annotations
from dataclasses import dataclass, field
from abc import ABC
from uuid import UUID

import pandas as pd

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.Enums.HowEnum import OperationEnum

from ResSimpy.DataModelBaseClasses.Well import Well
from typing import Sequence, Optional

from ResSimpy.Time.ISODateTime import ISODateTime


@dataclass(kw_only=True)
class Wells(ABC):
    _wells: Sequence[Well] = field(default_factory=list)
    _wells_loaded: bool = False

    def __init__(self, assume_loaded: bool = False) -> None:
        """Initialises the Wells class.

        Args:
            assume_loaded (bool): whether the class should assume that the Wells have already been loaded.
        """
        self._wells_loaded = assume_loaded
        self._wells = []

    @property
    def wells(self) -> Sequence[Well]:
        """Writes a list of well objects.

        Returns:
               A sequence ( usually of type List) of wells. If wells were not previously loaded,
               they will be loaded (using the _load before returning).
        """
        if not self._wells_loaded:
            self._load()
        return self._wells

    def _load(self) -> None:
        raise NotImplementedError("Implement this in the derived class")

    def get_all(self) -> Sequence[Well]:
        """Returns a sequence of wells usually of type list."""
        raise NotImplementedError("Implement this in the derived class")

    def get(self, well_name: str) -> Optional[Well]:
        """Gets a well by name.

        Args:
            well_name (str): The well name.
        """
        raise NotImplementedError("Implement this in the derived class")

    def get_df(self) -> pd.DataFrame:
        """Returns wells as pandas data frame."""
        raise NotImplementedError("Implement this in the derived class")

    def modify(self, well_name: str, completion_properties_list: list[DataObjectMixinDictType],
               how: OperationEnum = OperationEnum.ADD) -> None:
        """Modifies a completion in a named well using a list of properties. How enum determines if the completions are
        added, removed or modified.

        Args:
            well_name (str) : The well name.
            completion_properties_list (list[dict[str, None | float | int | str]]) : List of completions property
            dictionaries to pass through to each of the ADD/REMOVE/MODIFY operations.
            how (OperationEnum) : Check the OperationEnum for valid options.

        Returns:
            None
        """
        raise NotImplementedError("Implement this in the derived class")

    def add_completion(self, well_name: str, completion_properties: DataObjectMixinDictType,
                       preserve_previous_completions: bool = True, comments: Optional[str] = None) -> None:
        """Adds completion to an existing wellspecfile.

        Args:
            well_name(str): The well name.
            completion_properties(dict[str, None | float | int | str]): Properties of the added completion.
            preserve_previous_completions(bool): If true, a new perforation on a TIME card will keep previous
            completions from the nearest TIME card along with the new completion.
            comments(Optional[str]): Comments about the added completion.

        Returns:
            None
        """
        raise NotImplementedError("Implement this in the derived class")

    def remove_completion(self, well_name: str,
                          completion_properties: Optional[DataObjectMixinDictType] = None,
                          completion_id: Optional[UUID] = None) -> None:
        """Well name to remove the completion from.

        Args:
            well_name(str): Well name.
            completion_properties(Optional[dict[str, None | float | int | str]]): Completion properties.
            completion_id(Optional[UUID]): Completion unique identifier.

        Returns:
            None
        """
        raise NotImplementedError("Implement this in the derived class")

    def modify_completion(self, well_name: str, properties_to_modify: DataObjectMixinDictType,
                          completion_to_change: Optional[DataObjectMixinDictType] = None,
                          completion_id: Optional[UUID] = None,
                          comments: Optional[str] = None) -> None:
        """Well name to modify completion from.

        Args:
            well_name (str): Well name.
            properties_to_modify(dict[str, None | float | int | str]): Well properties to be modified.
            completion_to_change(optional[dict[str, None | float | int | str]]): Completion that will be modified.
            completion_id(Optional[UUID]): Completion unique identifier.
            comments(Optional[str]): Comments about the modified completion.

        Returns:
            None
        """
        raise NotImplementedError("Implement this in the derived class")

    def get_wells_overview(self) -> str:
        """Returns well information as a string format and updates the overview variable."""
        overview: str = ''
        for well in self.wells:
            overview += well.printable_well_info

        return overview

    def get_wells_dates(self) -> set[str]:
        """Returns a set of the unique dates in the wellspec file over all wells."""
        set_dates: set[str] = set()
        for well in self.wells:
            set_dates.update(set(well.dates_of_completions))

        return set_dates

    def get_wells_iso_dates(self) -> set[ISODateTime]:
        """Returns a set of the unique ISO dates in the wellspec file over all wells."""
        set_dates: set[ISODateTime] = set()
        for well in self.wells:
            set_dates.update({x.iso_date for x in well.completions})

        return set_dates

    @property
    def table_header(self) -> str:
        """Returns table header as a string."""
        raise NotImplementedError("Implement this in the derived class")

    @property
    def table_footer(self) -> str:
        """Returns table footer as a string."""
        raise NotImplementedError("Implement this in the derived class")



from abc import ABC
from dataclasses import dataclass, field
from typing import Sequence

from ResSimpy.DataModelBaseClasses.LGR import LGR


@dataclass
class LGRs(ABC):
    """Abstract base class for Local Grid Refinements (LGR)."""
    _lgr_list: Sequence[LGR] = field(default_factory=list)

    def __init__(self, lgr_list: None | Sequence[LGR] = None) -> None:
        """Initializes the LGRs class."""
        self._lgr_list: Sequence[LGR] = [] if lgr_list is None else lgr_list

    @property
    def lgr_list(self) -> Sequence[LGR]:
        """Returns the list of LGRs."""
        return self._lgr_list

    def get_all(self) -> Sequence[LGR]:
        """Returns all LGRs."""
        return self._lgr_list



from __future__ import annotations
from dataclasses import dataclass, field
import os
from typing import Optional, MutableMapping

from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.DataModels.FcsFile import FcsNexusFile
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.DataModels.NexusGasliftMethod import NexusGasliftMethod
from ResSimpy.DataModelBaseClasses.Gaslift import Gaslift
from ResSimpy.Utils.dynamic_method_manipulations import add_dynamic_method


@dataclass(kw_only=True)
class NexusGasliftMethods(Gaslift):
    """Class for collection of Nexus gaslift methods.

    Attributes:
        inputs (dict[int, NexusGasliftMethod]): Collection of Nexus gaslift methods, as a dictionary
        files (dict[int, NexusFile]): Dictionary collection of gaslift files, as defined in Nexus fcs file.
    """

    __inputs: MutableMapping[int, NexusGasliftMethod]
    __files: dict[int, NexusFile]
    __properties_loaded: bool = False  # Used in lazy loading
    __model_unit_system: UnitSystem
    _model_files: Optional[FcsNexusFile] = field(default=None, repr=False, compare=False)

    def __init__(self, model_unit_system: UnitSystem,
                 inputs: Optional[MutableMapping[int, NexusGasliftMethod]] = None,
                 files: Optional[dict[int, NexusFile]] = None, assume_loaded: bool = False,
                 model_files: Optional[FcsNexusFile] = None) -> None:
        """Initialises the NexusGasliftMethods class.

        Args:
            model_unit_system (UnitSystem): Unit system used in the model.
            inputs (Optional[MutableMapping[int, NexusGasliftMethod]]): Collection of Nexus gaslift methods.
            files (Optional[dict[int, NexusFile]]): Collection of gaslift files, as defined in Nexus fcs file.
            Keyed by the method number.
            assume_loaded (bool): If True, assumes that the gaslift methods are already loaded.
            model_files (Optional[FcsNexusFile]): FcsNexusFile object containing the whole set of Nexus files.
        """
        if inputs:
            self.__inputs = inputs
        else:
            self.__inputs: MutableMapping[int, NexusGasliftMethod] = {}
        if files:
            self.__files = files
        else:
            self.__files = {}
        self.__model_unit_system = model_unit_system
        self.__properties_loaded = assume_loaded
        self._model_files = model_files
        super().__init__()

    def __repr__(self) -> str:
        """Pretty printing gaslift methods."""
        if not self.__properties_loaded:
            self.load_gaslift_methods()
        printable_str = ''
        for table_num in self.__inputs.keys():
            printable_str += '\n--------------------------------\n'
            printable_str += f'GASLIFT method {table_num}\n'
            printable_str += '--------------------------------\n'
            printable_str += self.__inputs[table_num].__repr__()
            printable_str += '\n'

        return printable_str

    @property
    def inputs(self) -> MutableMapping[int, NexusGasliftMethod]:
        """Returns mapping of gaslift where keys are of type int.
        If gaslift properties are not loaded, it will load them.
        """
        if not self.__properties_loaded:
            self.load_gaslift_methods()
        return self.__inputs

    @property
    def files(self) -> dict[int, NexusFile]:
        """Returns dictionary of Nexus files where keys are of type int."""
        return self.__files

    def load_gaslift_methods(self) -> None:
        """Loads a collection of gaslift method properties from files.
        This method checks if gaslift files are available and reads their properties into the
        'NexusGasliftMethod' objects.
        """
        # Read in gaslift properties from Nexus gaslift method files
        if self.__files is not None and len(self.__files) > 0:  # Check if gaslift files exist
            for table_num in self.__files.keys():  # For each gaslift property method
                gaslift_file = self.__files[table_num]
                if gaslift_file.location is None:
                    raise ValueError(f'Unable to find gaslift file: {gaslift_file}')
                if os.path.isfile(gaslift_file.location):
                    # Create NexusGasliftMethod object
                    self.__inputs[table_num] = NexusGasliftMethod(file=gaslift_file, input_number=table_num,
                                                                  model_unit_system=self.__model_unit_system)
                    self.__inputs[table_num].read_properties()  # Populate object with gaslift props in file
        self.__properties_loaded = True

    @property
    def model_unit_system(self) -> UnitSystem:
        """Return the model unit system."""
        return self.__model_unit_system

    @property
    def keyword(self) -> str:
        """Return the keyword for gaslift methods."""
        return 'GASLIFT'

    def add_method(self, method: DynamicProperty, new_file_name: str, create_new_file: bool = False) -> None:
        """Adds a new gaslift method to the collection.

        Args:
            method (DynamicProperty): The gaslift method to add.
            new_file_name (str): The name of the file to save the method to.
            create_new_file (bool): Whether to create a new file for the method.
        """
        add_dynamic_method(dynamic_method_collection=self, method=method, new_file_name=new_file_name,
                           create_new_file=create_new_file)

    def _method_type(self) -> type[NexusGasliftMethod]:
        """Returns the expected type of the dynamic property."""
        return NexusGasliftMethod



"""Collection of Enumerations for penetration direction used in ResSimpy."""
from enum import Enum


class PenetrationDirectionEnum(str, Enum):
    """Enum representing the penetration direction options in other simulators."""
    X = 'X'
    Y = 'Y'
    Z = 'Z'
    FX = 'FX'
    FY = 'FY'



"""Class for handling drills in the Nexus Network.

This class is used to store and manipulate the wellheads in a NexusNetwork. It is stored as an instance in the
NexusNetwork class as "drill_sites". In Nexus this is the DRILLSITE table.
"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, TYPE_CHECKING
from uuid import UUID

import pandas as pd

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.FileOperations.File import File
from ResSimpy.GenericContainerClasses.DrillSites import DrillSites
from ResSimpy.Nexus.DataModels.Network.NexusDrillSite import NexusDrillSite
from ResSimpy.Nexus.nexus_add_new_object_to_file import AddObjectOperations
from ResSimpy.Nexus.nexus_collect_tables import collect_all_tables_to_objects
from ResSimpy.Nexus.nexus_modify_object_in_file import ModifyObjectOperations
from ResSimpy.Nexus.nexus_remove_object_from_file import RemoveObjectOperations
from ResSimpy.Utils.obj_to_dataframe import obj_to_dataframe

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import NexusNetwork


@dataclass
class NexusDrillSites(DrillSites):
    _drill_sites: list[NexusDrillSite] = field(default_factory=list)

    def __init__(self, parent_network: NexusNetwork) -> None:
        """Initialises the NexusDrillSites class.

        Args:
            parent_network (NexusNetwork): The network that the drill sites are a part of.
        """
        self.__parent_network: NexusNetwork = parent_network
        self._drill_sites: list[NexusDrillSite] = []
        self.__add_object_operations = AddObjectOperations(NexusDrillSite, self.table_header, self.table_footer,
                                                           self.__parent_network.model)
        self.__remove_object_operations = RemoveObjectOperations(self.__parent_network, self.table_header,
                                                                 self.table_footer)
        self.__modify_object_operations = ModifyObjectOperations(self)

    def get_all(self) -> list[NexusDrillSite]:
        """Returns a list of drills loaded from the simulator."""
        self.__parent_network.get_load_status()
        return self._drill_sites

    def get_by_name(self, name: str) -> Optional[NexusDrillSite]:
        """Returns a single drill site with the provided name loaded from the simulator.

        Args:
            name (str): name of the requested drill site
        Returns:
            NexusDrillSite: which has the same name as requested
        """
        to_return = filter(lambda x: False if x.name is None else x.name.upper() == name.upper(),
                           self._drill_sites)
        return next(to_return, None)

    def get_df(self) -> pd.DataFrame:
        """Returns drill site data as a dataframe."""
        return obj_to_dataframe(self._drill_sites)

    def load(self, surface_file: File, start_date: str, default_units: UnitSystem) -> None:
        """Loads drill site data from surface file.

        Args:
            surface_file(file): Nexus file representation of the surface file.
            start_date(str): start date of the run.
            default_units(UnitSystem): default units used if no units are found.
        """
        new_drill_sites, _ = collect_all_tables_to_objects(surface_file, {'DRILLSITE': NexusDrillSite},
                                                           start_date=start_date,
                                                           default_units=default_units,
                                                           date_format=self.__parent_network.model.date_format)
        drill_sites_list = new_drill_sites.get('DRILLSITE')
        self._add_to_memory(drill_sites_list)

    def _add_to_memory(self, additional_list: Optional[list[NexusDrillSite]]) -> None:
        """Extends the drill object by a list of drill sites provided to it.

        Args:
            additional_list (list[NexusDrill]): list of nexus drill sites to add to the drill list.
        """
        if additional_list is None:
            return
        self._drill_sites.extend(additional_list)

    def remove(self, obj_to_remove: DataObjectMixinDictType | UUID) -> None:
        """Remove a drill site from the network based on the properties matching a dictionary or id.

        Args:
            obj_to_remove (UUID | dict[str, None | str | float | int]): UUID of the drill site to remove or a dictionary
            with sufficient matching parameters to uniquely identify a drill site

        """
        self.__remove_object_operations.remove_object_from_network_main(
            obj_to_remove, self._network_element_name, self._drill_sites)

    def add(self, obj_to_add: DataObjectMixinDictType) -> None:
        """Adds a drill site to a network, taking a dictionary with properties for the new drill site.

        Args:
            obj_to_add (dict[str, None | str | float | int]): dictionary taking all the properties for the new
            drill.
            Requires date and a name.
        """
        new_object = self.__add_object_operations.add_network_obj(obj_to_add, NexusDrillSite, self.__parent_network)
        self._add_to_memory([new_object])

    def modify(self, obj_to_modify: DataObjectMixinDictType, new_properties: DataObjectMixinDictType) -> None:
        """Modifies an existing drill site based on a matching dictionary of properties.

        Partial matches allowed if precisely 1 matching node is found.
        Updates the properties with properties in the new_properties dictionary.

        Args:
            obj_to_modify (dict[str, None | str | float | int]): dictionary containing attributes to match in the
            existing drill sites.
            new_properties (dict[str, None | str | float | int]): properties to switch to in the new drill site.
        """
        self.__parent_network.get_load_status()
        self.__modify_object_operations.modify_network_object(obj_to_modify, new_properties,
                                                              self.__parent_network)

    @property
    def table_header(self) -> str:
        """Start of the drill site definition table."""
        return 'DRILLSITE'

    @property
    def table_footer(self) -> str:
        """End of the drill site definition table."""
        return 'END' + self.table_header

    def get_overview(self) -> str:
        """Returns overview of the drill site."""
        raise NotImplementedError('To be implemented')



"""This module contains the abstract base class for a collection of equilibration inputs."""
from dataclasses import dataclass
from typing import Mapping

from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.DataModelBaseClasses.DynamicPropertyContainer import DynamicPropertyContainer


@dataclass(kw_only=True)
class Equilibration(DynamicPropertyContainer):
    """The abstract base class for a collection of equilibration inputs.

    Attributes:
        inputs (dict[int, DynamicProperty]): Collection of equilibration inputs, as a dictionary.
    """

    @property
    def summary(self) -> str:
        """Returns string summary of Equilibration properties."""
        raise NotImplementedError("Implement this in derived class")

    @property
    def inputs(self) -> Mapping[int, DynamicProperty]:
        """Returns a mapping of equilibration inputs."""
        raise NotImplementedError("Implement this in the derived class")

    def to_string(self) -> str:
        """Create string with equilibration data in Nexus file format."""
        raise NotImplementedError('Implement this in the derived class.')



from enum import Enum


class PvtType(str, Enum):
    """Enum representing the different PVT types."""
    BLACKOIL = 'BLACKOIL'
    GASWATER = 'GASWATER'
    WATEROIL = 'WATEROIL'
    EOS = 'EOS'
    API = 'API'


class SeparatorType(str, Enum):
    """Enum representing the different separator types."""
    BLACKOIL = 'BLACKOIL'
    GASPLANT = 'GASPLANT'
    EOS = 'EOS'


class PhaseType(str, Enum):
    """Enum representing the different phase types."""
    OIL = 'OIL'
    LIQUID = 'LIQ'
    GAS = 'GAS'
    RESERVOIR = 'RES'
    COMBINED = 'COMB'
    NONE = 'NONE'



"""Class for collection of Nexus PVT property methods."""

from __future__ import annotations
from dataclasses import dataclass, field
import os
from typing import Optional, MutableMapping

from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.DataModels.FcsFile import FcsNexusFile
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.DataModels.NexusPVTMethod import NexusPVTMethod
from ResSimpy.DataModelBaseClasses.PVT import PVT
from ResSimpy.Utils.dynamic_method_manipulations import add_dynamic_method


@dataclass(kw_only=True)
class NexusPVTMethods(PVT):
    """Class for collection of Nexus PVT property methods.

    Attributes:
        inputs (dict[int, NexusPVTMethod]): Collection of Nexus PVT property methods, as a dictionary
        files (dict[int, NexusFile]): Dictionary collection of PVT property files, as defined in Nexus fcs file.
    """

    __inputs: MutableMapping[int, NexusPVTMethod]
    __files: dict[int, NexusFile]
    __properties_loaded: bool = False  # Used in lazy loading
    __model_unit_system: UnitSystem
    _model_files: Optional[FcsNexusFile] = field(default=None, repr=False, compare=False)

    def __init__(self, model_unit_system: UnitSystem, inputs: Optional[MutableMapping[int, NexusPVTMethod]] = None,
                 files: Optional[dict[int, NexusFile]] = None, assume_loaded: bool = False,
                 model_files: Optional[FcsNexusFile] = None) -> None:
        """Initialises the NexusPVTMethods class.

        Args:
            model_unit_system (UnitSystem): Unit system used in the model.
            inputs (Optional[MutableMapping[int, NexusPVTMethod]]): Collection of Nexus PVT property methods.
            files (Optional[dict[int, NexusFile]]): Collection of PVT property files, as defined in Nexus fcs file.
                Keyed by the method number.
            assume_loaded (bool): If True, assumes that the properties are already loaded.
            model_files (Optional[FcsNexusFile]): The Nexus fcs file containing the model files.
        """
        if inputs:
            self.__inputs = inputs
        else:
            self.__inputs: MutableMapping[int, NexusPVTMethod] = {}
        if files:
            self.__files = files
        else:
            self.__files = {}
        self.__model_unit_system = model_unit_system
        self.__properties_loaded = assume_loaded
        self._model_files = model_files
        super().__init__()

    def __repr__(self) -> str:
        """Pretty printing pvt methods."""
        if not self.__properties_loaded:
            self.load_pvt_methods()
        printable_str = ''
        for table_num in self.__inputs.keys():
            printable_str += '\n--------------------------------\n'
            printable_str += f'PVT method {table_num}\n'
            printable_str += '--------------------------------\n'
            printable_str += self.__inputs[table_num].__repr__()
            printable_str += '\n'

        return printable_str

    @property
    def inputs(self) -> MutableMapping[int, NexusPVTMethod]:
        """Returns mapping of Nexus PVT property methods.
        If properties are not loaded, it will load them first.
        """
        if not self.__properties_loaded:
            self.load_pvt_methods()
        return self.__inputs

    @property
    def files(self) -> dict[int, NexusFile]:
        """Returns dictionary of Nexus files."""
        return self.__files

    def load_pvt_methods(self) -> None:
        """Loads a collection of pvt properties from Nexus pvt method files."""
        # Read in pvt properties from Nexus pvt method files
        if self.__files is not None and len(self.__files) > 0:  # Check if pvt files exist
            for table_num in self.__files.keys():  # For each pvt property method
                pvt_file = self.__files[table_num]
                if pvt_file.location is None:
                    raise ValueError(f'Unable to find pvt file: {pvt_file}')
                if os.path.isfile(pvt_file.location):
                    # Create NexusPVTMethod object
                    self.__inputs[table_num] = NexusPVTMethod(file=pvt_file, input_number=table_num,
                                                              model_unit_system=self.__model_unit_system)
                    self.__inputs[table_num].read_properties()  # Populate object with pvt properties in file
        self.__properties_loaded = True

    @property
    def model_unit_system(self) -> UnitSystem:
        """Return the model unit system."""
        return self.__model_unit_system

    @property
    def summary(self) -> str:
        """Returns a string summary of 'NexusPVTMethods' inputs in a dictionary."""
        pvt_summary = ''
        for key, value in self.inputs.items():
            pvt_summary += f'        {key}: {value.file.location}\n'
        return pvt_summary

    @property
    def keyword(self) -> str:
        """Returns the keyword for the NexusPVTMethods."""
        return 'PVT'

    def add_method(self, method: DynamicProperty, new_file_name: str, create_new_file: bool = False) -> None:
        """Adds a new rock method to the collection.

        Args:
            method (DynamicProperty): The rock method to add.
            new_file_name (str): The name of the file to save the method to.
            create_new_file (bool): Whether to create a new file for the method.
        """
        add_dynamic_method(dynamic_method_collection=self, method=method, new_file_name=new_file_name,
                           create_new_file=create_new_file)

    def _method_type(self) -> type[NexusPVTMethod]:
        """Returns the expected type of the dynamic property."""
        return NexusPVTMethod



"""Class to hold Nexus Aquifer properties."""
from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Union
import numpy as np
import pandas as pd
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.NexusKeywords.aquifer_keywords import AQUIFER_SINGLE_KEYWORDS, AQUIFER_TABLE_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.aquifer_keywords import AQUIFER_KEYWORDS_VALUE_FLOAT, AQUIFER_KEYWORDS_VALUE_INT
from ResSimpy.Nexus.NexusKeywords.aquifer_keywords import AQUIFER_KEYWORDS, AQUIFER_TYPE_KEYWORDS
from ResSimpy.Enums.UnitsEnum import UnitSystem, SUnits, TemperatureUnits
from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.Units.AttributeMappings.DynamicPropertyUnitMapping import AquiferUnits

from ResSimpy.Utils.factory_methods import get_empty_dict_union
import ResSimpy.Nexus.nexus_file_operations as nfo


@dataclass(kw_only=True, repr=False)  # Doesn't need to write an _init_, _eq_ methods, etc.
class NexusAquiferMethod(DynamicProperty):
    """Class to hold Nexus Aquifer properties.

    Attributes:
        file (NexusFile): Nexus aquifer properties file object
        input_number (int): Aquifer properties method number in Nexus fcs file
        properties (dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                    dict[str, Union[float, pd.DataFrame]]]]):
            Dictionary holding all properties for a specific aquifer properties method. Defaults to empty dictionary.
    """

    # General parameters
    file: NexusFile
    properties: dict[str, Union[str, int, float, Enum, list[str], np.ndarray, pd.DataFrame,
                                dict[str, Union[float, pd.DataFrame]]]] = \
        field(default_factory=get_empty_dict_union)
    unit_system: UnitSystem
    __hash__ = DynamicProperty.__hash__

    def __init__(self, file: NexusFile, input_number: int, model_unit_system: UnitSystem,
                 properties: Optional[dict[str, Union[str, int, float, Enum, list[str], np.ndarray, pd.DataFrame,
                                      dict[str, Union[float, pd.DataFrame]]]]] = None) -> None:
        """Initialises the NexusAquiferMethod class.

        Args:
            file (NexusFile): NexusFile object associated with the aquifer method.
            input_number (int): method number for the aquifer method.
            model_unit_system (UnitSystem): unit system used in the model.
            properties: dictionary of properties for the aquifer method. Defaults to None.
        """
        if properties is not None:
            self.properties = properties
        else:
            self.properties = {}
        self.unit_system = model_unit_system
        super().__init__(input_number=input_number, file=file)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords: dict[str, tuple[str, type]] = {
            'BAQ': ('carter_tracy_constant', float),
            'CT': ('total_compressibility', float),
            'POR': ('porosity', float),
            'H': ('thickness', float),
            'RO': ('radius_to_inner_perimeter', float),
            'RE': ('radius_to_exterior_perimeter', float),
            'S': ('fraction_of_circular_boundary', float),
            'W': ('linear_aquifer_width', float),
            'L': ('linear_aquifer_length', float),
            'TC': ('time_conversion_factor', float),
            'VISC': ('viscosity', float),
            'K': ('permeability', float),
            'PAQI': ('initial_aquifer_pressure', float),
            'DAQI': ('datum_depth', float),
            'PI': ('productivity_index', float),
            'WEI': ('initial_encroachable_water_volume', float),
            'WAQI': ('initial_aquifer_volume', float),
        }
        return keywords

    @property
    def units(self) -> AquiferUnits:
        """Returns the attribute to unit map for the Aquifer method."""
        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']
        return AquiferUnits(unit_system=self.unit_system)

    def to_string(self) -> str:
        """Create string with aquifer data in Nexus file format."""
        printable_str = ''
        aquifer_dict = self.properties
        for key, value in aquifer_dict.items():
            if isinstance(value, pd.DataFrame):
                printable_str += f'{key}\n'
                printable_str += value.to_string(na_rep='', index=False) + '\n'
                if key == 'TRACER':
                    printable_str += 'ENDTRACER\n'
                printable_str += '\n'
            elif isinstance(value, Enum):
                if isinstance(value, UnitSystem) or isinstance(value, TemperatureUnits):
                    printable_str += f'{value.value}\n'
                elif isinstance(value, SUnits):
                    printable_str += f'SUNITS {value.value}\n'
            elif key == 'DESC' and isinstance(value, list):
                for desc_line in value:
                    printable_str += 'DESC ' + desc_line + '\n'
            elif key not in ['DESC']:
                if value == '':
                    printable_str += f'{key}\n'
                else:
                    printable_str += f'{key} {value}\n'
        printable_str += '\n'
        return printable_str

    def read_properties(self) -> None:
        """Read Nexus aquifer file contents and populate the NexusValveMethod object."""
        file_as_list = self.file.get_flat_list_str_file

        # Check for common input data
        nfo.check_for_and_populate_common_input_data(file_as_list, self.properties)

        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']

        # Initialize flags and containers to use to record properties as we iterate through aquifer file contents
        # Dictionary to record start and ending indices for tables
        aquifer_table_indices: dict[str, list[int]] = {}
        # Dictionary of flags indicating which tables are being read
        table_being_read: dict[str, bool] = {}
        for table_name in AQUIFER_TABLE_KEYWORDS:
            table_being_read[table_name] = False

        line_indx = 0
        for line in file_as_list:

            # Find standalone aquifer keywords, such as CARTER_TRACY or LINEAR
            if [i for i in line.split() if i in AQUIFER_TYPE_KEYWORDS + AQUIFER_SINGLE_KEYWORDS]:
                for word in AQUIFER_TYPE_KEYWORDS + AQUIFER_SINGLE_KEYWORDS:
                    if nfo.check_token(word, line):
                        self.properties[word] = ''
            # Find AQUIFER key-float value pairs, such as LINFAC 0.8 or BAQ 20.
            if [i for i in line.split() if i in AQUIFER_KEYWORDS_VALUE_FLOAT]:
                for key in AQUIFER_KEYWORDS_VALUE_FLOAT:
                    if nfo.check_token(key, line):
                        self.properties[key] = float(nfo.get_expected_token_value(key, line, file_as_list))
            # Find AQUIFER key-int value pairs, such as ITDPD 1 or IWATER 2
            if [i for i in line.split() if i in AQUIFER_KEYWORDS_VALUE_INT]:
                for key in AQUIFER_KEYWORDS_VALUE_INT:
                    if nfo.check_token(key, line):
                        self.properties[key] = int(nfo.get_expected_token_value(key, line, file_as_list))

            # Find beginning and ending indices of tables
            for table_key in AQUIFER_TABLE_KEYWORDS:
                if (table_key == 'TRACER' and table_being_read[table_key] and nfo.check_token('END' + table_key,
                                                                                              line)) \
                        or (table_key == 'TDPD' and table_being_read[table_key] and
                            [i for i in line.split() if nfo.check_token(i, line) and i in AQUIFER_KEYWORDS]):
                    table_being_read[table_key] = False
                    aquifer_table_indices[table_key][1] = line_indx
                if nfo.check_token(table_key, line):
                    table_being_read[table_key] = True
                    aquifer_table_indices[table_key] = [line_indx + 1, len(file_as_list)]

            line_indx += 1

        # Read in tables
        for key in aquifer_table_indices.keys():
            self.properties[key] = nfo.read_table_to_df(file_as_list[
                                                        aquifer_table_indices[key][0]:aquifer_table_indices[key][1]])



"""Collection of Enumerations for Units used in ResSimpy."""
from enum import Enum


# Enum representing the Salinity Units
class SUnits(str, Enum):
    MEQ_ML = 'MEQ/ML'
    PPM = 'PPM'


class TemperatureUnits(str, Enum):
    FAHR = 'FAHR'
    CELSIUS = 'CELSIUS'
    RANKINE = 'RANKINE'
    KELVIN = 'KELVIN'


class UnitSystem(str, Enum):
    ENGLISH = 'ENGLISH'
    LAB = 'LAB'
    METRIC = 'METRIC'
    METKGCM2 = 'METKG/CM2'
    METBAR = 'METBAR'
    METRIC_ATM = 'PVT-M'
    UNDEFINED = 'UNDEFINED'



from __future__ import annotations
from abc import ABC
from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.DataModelBaseClasses.NetworkObject import NetworkObject
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Units.AttributeMappings.NetworkUnitMapping import NetworkUnits


@dataclass(repr=False)
class Target(NetworkObject, ABC):
    control_quantity: Optional[str] = None
    control_conditions: Optional[str] = None
    control_connections: Optional[str] = None
    control_method: Optional[str] = None
    calculation_method: Optional[str] = None
    calculation_conditions: Optional[str] = None
    calculation_connections: Optional[str] = None
    value: Optional[float] = None
    add_value: Optional[float] = None
    region: Optional[str] = None
    priority: Optional[int] = None
    minimum_rate: Optional[float] = None
    minimum_rate_no_shut: Optional[float] = None
    guide_rate: Optional[str] = None
    max_change_pressure: Optional[float] = None
    rank_dt: Optional[float] = None
    control_type: Optional[str] = None
    calculation_type: Optional[str] = None
    region_number: Optional[int] = None

    def __init__(self, properties_dict: Optional[DataObjectMixinDictType] = None,
                 date: Optional[str] = None, date_format: Optional[DateFormat] = None, start_date: Optional[str] = None,
                 unit_system: Optional[UnitSystem] = None, name: Optional[str] = None,
                 control_quantity: Optional[str] = None, control_conditions: Optional[str] = None,
                 control_connections: Optional[str] = None, control_method: Optional[str] = None,
                 calculation_method: Optional[str] = None, calculation_conditions: Optional[str] = None,
                 calculation_connections: Optional[str] = None, value: Optional[float] = None,
                 add_value: Optional[float] = None, region: Optional[str] = None, priority: Optional[int] = None,
                 minimum_rate: Optional[float] = None, minimum_rate_no_shut: Optional[float] = None,
                 guide_rate: Optional[str] = None, max_change_pressure: Optional[float] = None,
                 rank_dt: Optional[float] = None, control_type: Optional[str] = None,
                 calculation_type: Optional[str] = None, region_number: Optional[int] = None) -> None:
        """Initialises the Target class.

        Args:
            properties_dict (dict): dict of the properties to set on the object.
            date (Optional[str]): The date of the object.
            date_format (Optional[DateFormat]): The date format that the object uses.
            start_date (Optional[str]): The start date of the model. Required if the object uses a decimal TIME.
            unit_system (Optional[UnitSystem]): The unit system of the object e.g. ENGLISH, METRIC.
            name (Optional[str]): The name of the object.
            control_quantity(Optional[str]): Specifies control type for connections.
            control_conditions(Optional[str]): Column heading keyword indicating the conditions used for control.
            control_connections(Optional[str]): Specifies connections controlled to meet targets, using lists or single
            connections; constraints ensure targets are satisfied.
            control_method(Optional[str]): Indicates the method used to allocate the target constraint to controlling
            connections.
            calculation_method (Optional[str]): Specifies how the target rate is determined.
            calculation_conditions (Optional[str]): Indicates the conditions used to calculate the target rate.
            calculation_connections (Optional[str]): Indicates which connections contribute to the target calculation.
            value (Optional[float]): Indicates the value for the target rate.
            add_value (Optional[float]): Specifies an amount to be added to the target rate calculated by CALCMETHOD.
            region (Optional[str]): Specified a region name from REGDATA table.
            priority (Optional[int]): Column heading for target priority; lower integers indicate higher priority.
            minimum_rate (Optional[float]): Indicates the minimum rate for each connection. Applies to all CTRL
            unless an individual qmin is specified for a connections in a subsequent TGTCON table.
            minimum_rate_no_shut (Optional[float]): Indicates the minimum rate for each connection. Applies to all CTRL
            unless an individual qmin_noshut is specified for a connections in a subsequent TGTCON table.
            guide_rate (Optional[str]): Specifies the guide rate for each connection.
            max_change_pressure (Optional[float]): Specifies the maximum rate of change of region pressure versus time.
            rank_dt (Optional[float]): Specifies the minimum time change between reranking the connections.
            control_type (Optional[str]): Specifies which connections (or wells) should be included from CTRLCONS
            connections.
            calculation_type (Optional[str]): Specifies which connections (or wells) should be included from CALCONS
            connections.
            region_number (Optional[int]): Region numbers, to which function input options are to be applied.
        """
        self.control_quantity = control_quantity
        self.control_conditions = control_conditions
        self.control_connections = control_connections
        self.control_method = control_method
        self.calculation_method = calculation_method
        self.calculation_conditions = calculation_conditions
        self.calculation_connections = calculation_connections
        self.value = value
        self.add_value = add_value
        self.region = region
        self.priority = priority
        self.minimum_rate = minimum_rate
        self.minimum_rate_no_shut = minimum_rate_no_shut
        self.guide_rate = guide_rate
        self.max_change_pressure = max_change_pressure
        self.rank_dt = rank_dt
        self.control_type = control_type
        self.calculation_type = calculation_type
        self.region_number = region_number

        super().__init__(properties_dict=properties_dict, date=date, date_format=date_format, start_date=start_date,
                         unit_system=unit_system, name=name)

    @property
    def units(self) -> NetworkUnits:
        """Returns the attribute to unit map for the WellConnection."""
        return NetworkUnits(self.unit_system)



"""Base class object for storing data related to well and node constraints."""

from __future__ import annotations
from abc import ABC
from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixin
from ResSimpy.Units.AttributeMappings.ConstraintUnitMapping import ConstraintUnits


@dataclass(repr=False)
class Constraint(DataObjectMixin, ABC):
    """Base class object for storing data related to well and node constraints."""
    # TODO: Add docstrings for this class
    name: Optional[str] = None
    max_surface_oil_rate:  Optional[float] = None
    max_surface_gas_rate: Optional[float] = None
    max_surface_water_rate: Optional[float] = None
    max_surface_liquid_rate: Optional[float] = None
    max_reservoir_oil_rate: Optional[float] = None
    max_reservoir_gas_rate: Optional[float] = None
    max_reservoir_water_rate: Optional[float] = None
    max_reservoir_liquid_rate: Optional[float] = None
    min_surface_oil_rate: Optional[float] = None
    min_surface_gas_rate: Optional[float] = None
    min_surface_water_rate: Optional[float] = None
    min_surface_liquid_rate: Optional[float] = None
    min_reservoir_total_fluids_rate: Optional[float] = None
    bottom_hole_pressure: Optional[float] = None
    tubing_head_pressure: Optional[float] = None
    max_reservoir_total_fluids_rate: Optional[float] = None
    max_avg_comp_dp: Optional[float] = None
    max_comp_dp: Optional[float] = None

    # attributes below here are also used in other simulators
    # to determine proper workover strategy in WECON
    max_watercut: Optional[float] = None
    max_watercut_plug: Optional[float] = None
    max_watercut_plugplus: Optional[float] = None
    max_watercut_perf: Optional[float] = None
    max_watercut_perfplus: Optional[float] = None
    max_wor: Optional[float] = None
    max_wor_plug: Optional[float] = None
    max_wor_plug_plus: Optional[float] = None
    max_wor_perf: Optional[float] = None
    max_wor_perfplus: Optional[float] = None
    max_gor: Optional[float] = None
    max_gor_plug: Optional[float] = None
    max_gor_plug_plus: Optional[float] = None
    max_gor_perf: Optional[float] = None
    max_gor_perfplus: Optional[float] = None
    max_lgr: Optional[float] = None
    max_lgr_plug: Optional[float] = None
    max_lgr_plug_plus: Optional[float] = None
    max_lgr_perf: Optional[float] = None
    max_lgr_perfplus: Optional[float] = None
    convert_qmult_to_reservoir_barrels: Optional[bool] = None
    active_node: Optional[bool] = None
    gor_limit: Optional[float] = None

    @property
    def units(self) -> ConstraintUnits:
        """Returns the attribute to unit map for the constraint."""
        return ConstraintUnits(self.unit_system)



"""A library to help read and manipulate input decks for reservoir simulators.

.. autosummary::
    :toctree: _autosummary
    :caption: API Reference
    :template: custom-module-template.rst
    :recursive:

    Nexus
    Utils
    Enums
"""

__version__ = "2.4.12"  # Set at build time

from ResSimpy.Nexus.NexusSimulator import NexusSimulator
from ResSimpy.OpenGoSim.OpenGoSimSimulator import OpenGoSimSimulator

__all__ = [
    "NexusSimulator",
    "OpenGoSimSimulator",
   ]



from dataclasses import dataclass
from typing import Mapping

from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.DataModelBaseClasses.DynamicPropertyContainer import DynamicPropertyContainer


@dataclass(kw_only=True)
class RelPerm(DynamicPropertyContainer):
    """The abstract base class for a collection of relative permeability and capillary pressure property inputs.

    Attributes:
        inputs (dict[int, DynamicProperty]): Dictionary collection of relperm and capillary pressure property inputs.
    """

    @property
    def summary(self) -> str:
        """Returns string summary of RelPerm properties."""
        raise NotImplementedError("Implement this in the derived class")

    @property
    def inputs(self) -> Mapping[int, DynamicProperty]:
        """Returns mapping of dynamic property instance as an int."""
        raise NotImplementedError("Implement this in the derived class")



"""Class representing DrillSite Parameters in a Nexus Network."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.DataModelBaseClasses.DrillSite import DrillSite
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat


@dataclass(kw_only=True)
class NexusDrillSite(DrillSite):
    """Class representing DrillSite Parameters in a Nexus Network.

    Attributes:
        max_rigs (int):  The maximum number of rigs allowed at the drill site.
    """

    def __init__(self, unit_system: Optional[UnitSystem] = None, name: Optional[str] = None, date: Optional[str] = None,
                 date_format: Optional[DateFormat] = None, start_date: Optional[str] = None,
                 max_rigs: int = 1, properties_dict: Optional[DataObjectMixinDictType] = None)\
            -> None:
        """Initialises the NexusDrill class.

        Args:
            unit_system (Optional[UnitSystem]): The unit system of the object e.g. ENGLISH, METRIC.
            name (Optional[str]): The name of the well / will list the drill relates to.
            date (Optional[str]): The date of the object.
            date_format (Optional[DateFormat]): The date format that the object uses.
            start_date (Optional[str]): The start date of the model. Required if the object uses a decimal TIME.
            max_rigs (Optional[int]): The maximum number of rigs allowed at the drill site.
            properties_dict (dict): dict of the properties to set on the object.
        """

        super().__init__(unit_system=unit_system, name=name, date_format=date_format, start_date=start_date, date=date,
                         max_rigs=max_rigs, properties_dict=properties_dict)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        nexus_mapping = {
            'NAME': ('name', str),
            'MAXRIGS': ('max_rigs', int)
        }

        return nexus_mapping



import re


def check_if_string_is_float(x: str) -> bool:
    """Function to check if a string can be successfully cast to a float.

    Args:
        x (str): String input to check

    Returns:
        bool: True if string can be successfully cast to a float, otherwise false.
    """
    try:
        float(x)
        return True
    except ValueError:
        return False


def convert_to_number(x: str) -> int | float:
    """Function that takes a numerical string and casts it to an int or float, as appropriate.

    Args:
        x (str): Numerical string

    Returns:
        int | float: integer or float value from input numerical string
    """
    try:
        # Try to convert the string to an integer
        return int(x)
    except ValueError:
        # If it fails, try to convert the string to a float
        try:
            return float(x)
        except ValueError:
            raise ValueError(f'Provided string {x} is erroneous and needs to be either an integer or a float.')


def expand_string_list_of_numbers(s: str) -> str:
    """Function that's a string of numbers with repeats such as 1 2 3*4 5 and expands it appropriately into 1 2 4 4 4 5.

    Args:
        s (str): Input string of numbers with repeats, e.g., 1 2 3*4 5

    Returns:
        str: Expanded string of numbers, e.g., 1 2 4 4 4 5
    """
    # Find all occurrences of the pattern 'count*number' where number can be an integer or a float
    pattern = re.compile(r'(\d+)\*(\d+\.?\d*[eE]?[-+]?\d+)')
    matches = pattern.findall(s)

    # Replace each match with the expanded string
    for match in matches:
        count = int(match[0])
        str_value = match[1]
        value = convert_to_number(str_value)
        expanded = ' '.join([str(value)] * count)
        s = s.replace(f'{count}*{str_value}', expanded)

    return s


def is_number(s: str) -> bool:
    """Function that checks for floats, numbers and scientific notation to see if it can be converted to a float.

    Args:
        s (str): String input to check

    Returns:
        bool: True if string can be converted into float
    """

    try:
        float(s)
        return True
    except ValueError:
        return False



"""Class for handling well connections in the Nexus Network.

This class is used to store and manipulate the well
connections in a NexusNetwork. It is stored as an instance in the NexusNetwork class as "well_connections".
In Nexus this is the WELLS table.
"""
from __future__ import annotations
from dataclasses import dataclass, field
from typing import Optional, TYPE_CHECKING, Sequence
from uuid import UUID

import pandas as pd

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.FileOperations.File import File
from ResSimpy.Nexus.nexus_add_new_object_to_file import AddObjectOperations
from ResSimpy.Nexus.nexus_modify_object_in_file import ModifyObjectOperations
from ResSimpy.Nexus.nexus_remove_object_from_file import RemoveObjectOperations
from ResSimpy.Utils.obj_to_dataframe import obj_to_dataframe
from ResSimpy.Nexus.DataModels.Network.NexusWellConnection import NexusWellConnection
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.nexus_collect_tables import collect_all_tables_to_objects
from ResSimpy.GenericContainerClasses.WellConnections import WellConnections

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import NexusNetwork


@dataclass
class NexusWellConnections(WellConnections):
    """Class for handling well connections in the Nexus Network.

    This class is used to store and manipulate the well connections in a NexusNetwork. It is stored as an instance in
    the NexusNetwork class as "well_connections". In Nexus this is the WELLS table. The list of well connections can
    be accessed in the NexusNetwork class through the get_all method.
    """
    _well_connections: list[NexusWellConnection] = field(default_factory=list)

    def __init__(self, parent_network: NexusNetwork) -> None:
        """Initialises the NexusWellConnections class.

        Args:
            parent_network (NexusNetwork): The network that the well connections are a part of.
        """
        self.__parent_network: NexusNetwork = parent_network

        self.__add_object_operations = AddObjectOperations(NexusWellConnection, self.table_header, self.table_footer,
                                                           self.__parent_network.model)
        self.__remove_object_operations = RemoveObjectOperations(self.__parent_network, self.table_header,
                                                                 self.table_footer)
        self.__modify_object_operations = ModifyObjectOperations(self)
        super().__init__(parent_network=parent_network)

    def get_all(self) -> Sequence[NexusWellConnection]:
        """Returns a list of well connections loaded from the simulator."""
        self.__parent_network.get_load_status()
        return self._well_connections

    def get_by_name(self, name: str) -> Optional[NexusWellConnection]:
        """Returns a single well connection with the provided name loaded from the simulator.

        Args:
            name (str): name of the requested well connection

        Returns:
            NexusWellConnection: which has the same name as requested
        """
        self.__parent_network.get_load_status()
        to_return = filter(lambda x: False if x.name is None else x.name.upper() == name.upper(),
                           self._well_connections)
        return next(to_return, None)

    def get_df(self) -> pd.DataFrame:
        """Creates a dataframe representing all processed well connections data in a surface file.

        Returns:
            DataFrame: of the properties of the well connections through time with each row representing a single well \
            connection.
        """
        self.__parent_network.get_load_status()
        return obj_to_dataframe(self._well_connections)

    def get_overview(self) -> str:
        """Returns overview of the well connections."""
        raise NotImplementedError('To be implemented')

    def load(self, surface_file: File, start_date: str, default_units: UnitSystem) -> None:
        """Loads well connections surface file as Nexus.

        Args:
            surface_file(file): path to the surface file.
            start_date(str): start date of the run.
            default_units(UnitSystem): default units to use if no units are found.
        """
        new_well_connections, _ = collect_all_tables_to_objects(surface_file, {'WELLS': NexusWellConnection,
                                                                               'GASWELLS': NexusWellConnection},
                                                                start_date=start_date,
                                                                default_units=default_units,
                                                                date_format=self.__parent_network.model.date_format)
        cons_list = new_well_connections.get('WELLS')
        gas_cons_list = new_well_connections.get('GASWELLS')
        self._add_to_memory(cons_list)
        self._add_to_memory(gas_cons_list)

    def remove(self, obj_to_remove: DataObjectMixinDictType | UUID) -> None:
        """Remove a wellbore from the network based on the properties matching a dictionary or id.

        Args:
            obj_to_remove (UUID | DataObjectMixinDictType): UUID of the wellbore to remove or a dictionary \
            with sufficient matching parameters to uniquely identify a wellbore

        """
        self.__remove_object_operations.remove_object_from_network_main(
            obj_to_remove, self._network_element_name, self._well_connections)

    def add(self, obj_to_add: DataObjectMixinDictType) -> None:
        """Adds a well connection to a network, taking a dictionary with properties for the new well connection.

        Args:
            obj_to_add (DataObjectMixinDictType): dictionary taking all the properties for the new
            well connection.
            Requires date and a name.
        """
        new_object = self.__add_object_operations.add_network_obj(obj_to_add, NexusWellConnection,
                                                                  self.__parent_network)
        self._add_to_memory([new_object])

    def modify(self, obj_to_modify: DataObjectMixinDictType,
               new_properties: DataObjectMixinDictType) -> None:
        """Modifies an existing well connection based on a matching dictionary of properties.

        Partial matches allowed if precisely 1 matching node is found.
        Updates the properties with properties in the new_properties dictionary.

        Args:
            obj_to_modify (DataObjectMixinDictType): dictionary containing attributes to match in the
            existing well connections.
            new_properties (DataObjectMixinDictType): properties to switch to in the new well connection
        """
        self.__parent_network.get_load_status()
        self.__modify_object_operations.modify_network_object(obj_to_modify, new_properties,
                                                              self.__parent_network)

    @property
    def table_header(self) -> str:
        """Returns the header for the wellconnections table in surface file."""
        return 'WELLS'

    @property
    def table_footer(self) -> str:
        """Returns the footer of the wellconnections table in surface file."""
        return 'END' + self.table_header

    @property
    def well_connections(self) -> Sequence[NexusWellConnection]:
        """Returns a list of all well connections loaded."""
        self.__parent_network.get_load_status()
        return self._well_connections



"""Abstract base class for holding the network data for holding all components of the production networks."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, Tuple, Sequence, Any, Mapping

from ResSimpy.DataModelBaseClasses.NodeConnection import NodeConnection
from ResSimpy.GenericContainerClasses.Constraints import Constraints
from ResSimpy.GenericContainerClasses.NetworkLists import NetworkLists
from ResSimpy.GenericContainerClasses.NodeConnections import NodeConnections
from ResSimpy.GenericContainerClasses.Nodes import Nodes
from ResSimpy.GenericContainerClasses.Targets import Targets
from ResSimpy.GenericContainerClasses.WellConnections import WellConnections
from ResSimpy.GenericContainerClasses.Wellbores import Wellbores
from ResSimpy.GenericContainerClasses.Wellheads import Wellheads


@dataclass(kw_only=True, init=False)
class Network(ABC):
    """Abstract base class for holding the network data, holding all components of the production networks."""
    nodes: Optional[Nodes]
    connections: Optional[NodeConnections]
    constraints: Optional[Constraints]
    targets: Optional[Targets] = None
    well_connections: Optional[WellConnections] = None
    wellheads: Optional[Wellheads] = None
    wellbores: Optional[Wellbores] = None
    welllists: Optional[NetworkLists] = None
    _has_been_loaded: bool = False

    def __init__(self, nodes: Optional[Nodes] = None, connections: Optional[NodeConnections] = None,
                 constraints: Optional[Constraints] = None, targets: Optional[Targets] = None,
                 well_connections: Optional[WellConnections] = None, wellheads: Optional[Wellheads] = None,
                 wellbores: Optional[Wellbores] = None, welllists: Optional[NetworkLists] = None,
                 assume_loaded: bool = False) -> None:
        """Initialising the Network Abstract Base class."""
        self._has_been_loaded = assume_loaded
        self.nodes = nodes
        self.connections = connections
        self.constraints = constraints
        self.well_connections = well_connections
        self.targets = targets
        self.wellbores = wellbores
        self.wellheads = wellheads
        self.welllists = welllists

    @abstractmethod
    def load(self) -> None:
        """Loads Network data."""
        raise NotImplementedError("Implement this in the derived class")

    def get_load_status(self) -> bool:
        """Checks load status and loads the network if it hasn't already been loaded."""
        if not self._has_been_loaded:
            self.load()
        return self._has_been_loaded

    def get_connected_objects(self, connection_name: str) -> Tuple[list[NodeConnection], list[NodeConnection]]:
        """Returns a list of all objects connected to a node connection in a surface network.

        Arguments:
            connection_name (str): The name of the connection to retrieve the connected nodes for.
        """

        if self.connections is None:
            raise ValueError("No connections found for this model.")

        connection_to_check = self.connections.get_by_name(name=connection_name)

        if connection_to_check is None or not isinstance(connection_to_check, NodeConnection):
            raise ValueError(f"Invalid node connection requested:{connection_name}")

        all_connections = self.connections.get_all()
        connections_directly_before_connection = [x for x in all_connections if
                                                  x.node_out == connection_name and x.name != connection_name]

        # Add the in connection directly referenced by the connection itself if it is missing
        if connection_to_check.node_in != connection_name:
            node_in_connection = self.connections.get_by_name(name=connection_to_check.node_in)
            if node_in_connection is not None and node_in_connection not in connections_directly_before_connection:
                connections_directly_before_connection.append(node_in_connection)

        connections_directly_after_connection = [x for x in all_connections if
                                                 x.node_in == connection_name and x.name != connection_name]

        # Add the out connection directly referenced by the connection itself if it is missing
        if connection_to_check.node_out != connection_name:
            node_out_connection = self.connections.get_by_name(name=connection_to_check.node_out)
            if node_out_connection is not None and node_out_connection not in connections_directly_after_connection:
                connections_directly_after_connection.append(node_out_connection)

        connections_in = self.__get_linked_connections(all_connections=all_connections,
                                                       direct_connections=connections_directly_before_connection,
                                                       search_upwards=True)

        # Reverse the order of the in connections to keep the order from 'top to bottom' in the network
        connections_in = connections_in[::-1]

        connections_out = self.__get_linked_connections(all_connections=all_connections,
                                                        direct_connections=connections_directly_after_connection,
                                                        search_upwards=False)

        return connections_in, connections_out

    def print_connected_objects(self, connection_name: str) -> str:
        """Prints a list of the connected objects.

        Arguments:
            connection_name (str): The name of the connection to retrieve the connected nodes for.
        """

        if self.connections is None:
            raise ValueError("No connections found for this model.")

        requested_node = self.connections.get_by_name(name=connection_name)
        if requested_node is None or not isinstance(requested_node, NodeConnection):
            raise ValueError(f"Invalid node connection requested:{connection_name}")

        connections_str = """NAME            TYPE
-----------------------
"""

        connected_objects = self.get_connected_objects(connection_name=connection_name)

        # Add the in node connections, then the object itself, then the out node connections
        for node in connected_objects[0]:
            connections_str += f"{node.name:<16}{node.con_type}\n"

        connections_str += f"\n{requested_node.name:<16}{requested_node.con_type} - Requested Node\n\n"

        for node in connected_objects[1]:
            connections_str += f"{node.name:<16}{node.con_type}\n"

        return connections_str

    def __get_linked_connections(self, all_connections: Sequence[NodeConnection],
                                 direct_connections: list[NodeConnection], search_upwards: bool) \
            -> list[NodeConnection]:
        """Returns a list of connections linked to the provided direct connections."""

        if self.connections is None:
            raise ValueError("No connections found for this model.")

        endpoints = ['GAS', 'SINK', 'WATER', 'IPR-SOURCE', 'IPR-SINK', 'FIELD']
        found_connections: list[NodeConnection] = []

        # Loop through all the direct connections, getting all the connections connected to each.
        for current_node in direct_connections:
            next_node_name = current_node.name
            while (next_node_name is not None
                   and next_node_name not in endpoints
                   and next_node_name not in [x.name for x in found_connections]):

                next_node = self.connections.get_by_name(name=next_node_name)
                if next_node is None or not isinstance(next_node, NodeConnection):
                    raise ValueError(f"Invalid node connection name found:{next_node_name}")

                found_connections.append(next_node)

                if search_upwards:
                    if next_node.node_in != next_node.name:
                        next_node_name = next_node.node_in
                    else:
                        # Find the connection linked to this node via 'node_out' on another connection instead
                        next_node_found = False
                        for connection in all_connections:
                            if connection.node_out == next_node.name:
                                next_node_name = connection.name
                                next_node_found = True
                                break

                        if not next_node_found:
                            break

                else:
                    if next_node.node_out != next_node.name:
                        next_node_name = next_node.node_out
                    else:
                        # Find the connection linked to this node via 'node_in' on another connection instead
                        next_node_found = False
                        for connection in all_connections:
                            if connection.node_in == next_node.name:
                                next_node_name = connection.name
                                next_node_found = True
                                break

                        if not next_node_found:
                            break

        return found_connections

    def get_all_linked_objects(self, object_name: str) -> dict[str, list[Any]]:
        """Gets a list of all the objects in the surface network linked to the specified object name.

        Args:
            object_name(str): The name of the object to search for.
        """
        linked_objects = {}
        linked_well_list_names = []
        linked_con_list_names = []
        linked_connection_names = []

        if self.nodes is not None:
            linked_nodes = [x for x in self.nodes.get_all() if x.name == object_name]
            linked_objects['NODES'] = linked_nodes

        if self.connections is not None:
            linked_connections = [x for x in self.connections.get_all() if x.name == object_name]
            linked_objects['CONNECTIONS'] = linked_connections
            linked_connection_names = [x.name for x in linked_connections]

        if self.well_connections is not None:
            linked_well_connections = [x for x in self.well_connections.get_all() if x.name == object_name]
            linked_objects['WELL_CONNECTIONS'] = linked_well_connections
            linked_well_connection_names = [x.name for x in linked_well_connections]
            linked_connection_names.append(linked_well_connection_names)

        if self.constraints is not None:
            if isinstance(self.constraints.constraints, Mapping):
                linked_constraints = list(self.constraints.constraints[object_name])
            else:
                linked_constraints = [x for x in self.constraints.constraints if x.name == object_name]
            linked_objects['CONSTRAINTS'] = linked_constraints

        if hasattr(self, 'conlists') and self.conlists is not None:
            linked_connection_lists = [x for x in self.conlists.lists if object_name in x.elements_in_the_list]
            linked_objects['CONLISTS'] = linked_connection_lists
            linked_con_list_names = [x.name for x in linked_connection_lists]

        if self.welllists is not None:
            linked_well_lists = [x for x in self.welllists.lists if object_name in x.elements_in_the_list]
            linked_objects['WELLLISTS'] = linked_well_lists
            linked_well_list_names = [x.name for x in linked_well_lists]

        if self.targets is not None:
            linked_targets = [x for x in self.targets.get_all() if x.control_connections in linked_well_list_names or
                              x.control_connections in linked_connection_names or
                              x.control_connections in linked_con_list_names]
            linked_objects['TARGETS'] = linked_targets

        if self.wellheads is not None:
            linked_well_heads = [x for x in self.wellheads.get_all() if x.well == object_name]
            linked_objects['WELLHEADS'] = linked_well_heads

        if self.wellbores is not None:
            linked_well_bores = [x for x in self.wellbores.get_all() if x.name == object_name]
            linked_objects['WELLBORES'] = linked_well_bores

        return linked_objects



"""Abstract base class for constraints."""
from __future__ import annotations
from dataclasses import dataclass, field
from abc import ABC, abstractmethod
import pandas as pd

from ResSimpy.DataModelBaseClasses.Constraint import Constraint
from typing import Optional, Mapping, Sequence

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType


@dataclass(kw_only=True)
class Constraints(ABC):
    """Abstract base class for constraints."""
    _constraints: Mapping[str, Sequence[Constraint]] | list[Constraint] = field(default_factory=dict)

    @property
    def constraints(self) -> Mapping[str, Sequence[Constraint]] | list[Constraint]:
        """Returns mapping of constraints."""
        return self._constraints

    @abstractmethod
    def get_all(self, object_name: Optional[str] = None, date: Optional[str] = None) \
            -> Mapping[str, Sequence[Constraint]]:
        """Get the constraints of the existing model with optional parameters to filter for name and date.

        Args:
            object_name(Optional[str] = None): name of the connection, node or wellname to return. Defaults to None.
            date(Optional[str] = None)): date in model format to filter the dates to in the constraints.
        """
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def get_df(self) -> pd.DataFrame:
        """Creates a dataframe representing all processed constraint data in a surface file."""
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def get_overview(self) -> str:
        """Returns an overview of the constraint."""
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def remove(self) -> None:
        """Remove a constraint."""
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def add(self,
            constraint_to_add: DataObjectMixinDictType | Constraint,
            comments: str | None,
            ) -> None:
        """Adds a constraint to the network and corresponding surface file.

        Args:
            constraint_to_add(dict[str, None | float | int | str | UnitSystem] | Constraint): properties of the
            constraints or a constraint object.
            comments(str): add optional post line comments.
        """
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def modify(self, name: str,
               current_constraint: DataObjectMixinDictType | Constraint,
               new_constraint_props: DataObjectMixinDictType | Constraint,
               comments: Optional[str] = None) \
            -> None:
        """Adds optional post line comments in the modified constraint.

        Args:
            name(str): The well name.
            current_constraint(DataObjectMixinDictType | Constraint):Modify an existing constraint.
            new_constraint_props(dict[str, None | float | int | str | UnitSystem] | Constraint): dictionary or
            constraint to update the constraint with.
            comments(Optional[str]):??
        """
        raise NotImplementedError("Implement this in the derived class")



"""Used to represent a connection between two nodes in a Nexus network."""
from __future__ import annotations
from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.DataModelBaseClasses.NodeConnection import NodeConnection


@dataclass(repr=False)
class NexusNodeConnection(NodeConnection):
    """Used to represent a connection between two nodes in a Nexus network.

    Attributes:
    hyd_method: hydraulic lift correlation method used (METHOD)
    pvt_method: pvt method number (IPVT)
    water_method: water method number (IWAT)
    bat_method: bat method number (IBAT)
    elevation_profile: elevation profile as a string (ELEVPR)
    measured_depth_in: measured depth of the input node (MDIN)
    measured_depth_out: measured depth of the output node (MDOUT)
    diameter: diameter of the connecting pipe (DIAMETER)
    inner_diameter: outer diameter of the pipe (INNERDIAM)
    roughness: roughness of the pipe (ROUGHNESS)
    heat_transfer_coeff: the heat transfer coefficient of the pipe (HTC)
    temperature_profile: temperature profile used for the connection. (TEMPPR)
    length: length of the connection (LENGTH)
    delta_depth: change in depth between node_in and node_out (DDEPTH)
    connection_number: used in place of name as a numbered connection (NUMBER)
    seawater_profile: seawater profile used. (SEAWPR)
    rate_mult: multiplier to the rate (RATEMULT)
    polymer: whether polymer is a stream here (POLYMER)
    dp_add: Additional delta pressure (DPADD).
    """

    bat_method: Optional[int] = None
    elevation_profile: Optional[str] = None
    measured_depth_in: Optional[float] = None
    measured_depth_out: Optional[float] = None
    diameter: Optional[float] = None
    inner_diameter: Optional[float] = None
    roughness: Optional[float] = None
    heat_transfer_coeff: Optional[float] = None
    temperature_profile: Optional[float] = None
    length: Optional[float] = None
    delta_depth: Optional[float] = None
    connection_number: Optional[int] = None
    seawater_profile: Optional[str] = None
    rate_mult: Optional[float] = None
    polymer: Optional[str] = None
    dp_add: Optional[float] = None
    dt_add: Optional[float] = None
    temperature_in: Optional[float] = None
    temperature_out: Optional[float] = None
    tracer: Optional[str] = None
    heat_conductivity: Optional[float] = None
    insulation_thickness: Optional[float] = None
    insulation_conductivity: Optional[float] = None
    gravity_pressure_gradient_mult: Optional[float] = None
    friction_pressure_gradient_mult: Optional[float] = None
    acceleration_pressure_gradient_mult: Optional[float] = None

    def __init__(self, properties_dict: None | DataObjectMixinDictType = None, date: Optional[str] = None,
                 date_format: Optional[DateFormat] = None, start_date: Optional[str] = None, name: Optional[str] = None,
                 unit_system: Optional[UnitSystem] = None, bat_method: Optional[int] = None,
                 elevation_profile: Optional[str] = None, measured_depth_in: Optional[float] = None,
                 measured_depth_out: Optional[float] = None, diameter: Optional[float] = None,
                 inner_diameter: Optional[float] = None, roughness: Optional[float] = None,
                 heat_transfer_coeff: Optional[float] = None, temperature_profile: Optional[float] = None,
                 length: Optional[float] = None, delta_depth: Optional[float] = None,
                 connection_number: Optional[int] = None, seawater_profile: Optional[str] = None,
                 rate_mult: Optional[float] = None, polymer: Optional[str] = None, dp_add: Optional[float] = None,
                 dt_add: Optional[float] = None, temperature_in: Optional[float] = None,
                 temperature_out: Optional[float] = None, tracer: Optional[str] = None,
                 heat_conductivity: Optional[float] = None, insulation_thickness: Optional[float] = None,
                 insulation_conductivity: Optional[float] = None,
                 gravity_pressure_gradient_mult: Optional[float] = None,
                 friction_pressure_gradient_mult: Optional[float] = None,
                 acceleration_pressure_gradient_mult: Optional[float] = None,
                 node_in: Optional[str] = None, node_out: Optional[str] = None, con_type: Optional[str] = None,
                 hyd_method: Optional[str | int] = None, pvt_method: Optional[int] = None,
                 water_method: Optional[int] = None) -> None:
        """Initialises the NexusNodeConnection class.

        Args:
            properties_dict (dict): dict of the properties to set on the object.
            date (Optional[str]): The date of the object.
            date_format (Optional[DateFormat]): The date format that the object uses.
            start_date (Optional[str]): The start date of the model. Required if the object uses a decimal TIME.
            unit_system (Optional[UnitSystem]): The unit system of the object e.g. ENGLISH, METRIC.
            name (Optional[str]): The name of the object.
            bat_method (Optional[float]): Bat method (IBAT).
            elevation_profile (Optional[str]): Elevation profile identifier (ELEVPR).
            measured_depth_in (Optional[float]): Measured depth of the input node (MDIN)
            measured_depth_out (Optional[float]): Measured depth of the output node (MDOUT).
            diameter (Optional[float]): diameter of the connecting pipe (DIAMETER).
            inner_diameter (Optional[float]): Outer diameter of the pipe (INNERDIAM).
            roughness (Optional[float]): Roughness of the pipe (ROUGHNESS)
            heat_transfer_coeff (Optional[float]): The heat transfer coefficient of the pipe (HTC).
            temperature_profile (Optional[float]): Temperature profile used for the connection (TEMPPR).
            length (Optional[float]): Length of the connection (LENGTH).
            delta_depth (Optional[float]): Change in depth between node_in and node_out (DDEPTH).
            connection_number (Optional[int]): Used in place of name as a numbered connection (NUMBER).
            seawater_profile (Optional[str]): Seawater profile used (SEAWPR).
            rate_mult (Optional[float]): Multiplier to the rate (RATEMULT).
            polymer (Optional[str]): Whether polymer is a stream here (POLYMER).
            dp_add (Optional[float]): Additional delta pressure (DPADD).
            dt_add (Optional[float]):  Additional temperature change (DTADD).
            temperature_in (Optional[float]): Temperature at nodein.
            temperature_out (Optional[float]): Temperature at node-out.
            tracer (Optional[str]): Specifies tracer concentrations.
            heat_conductivity (Optional[float]): Defines heat conductivity (HCOND).
            insulation_thickness (Optional[float]): Defines insulation thickness (INSTHN).
            insulation_conductivity (Optional[float]): Defines insulation conductivity (INSK).
            gravity_pressure_gradient_mult (Optional[float]): Correction Factor for the gravity pressure gradient for
            pressure drop correlations (GRPGCR).
            friction_pressure_gradient_mult (Optional[float]):  Correction Factor for the friction pressure gradient for
            pressure drop correlations (FRPGCR).
            acceleration_pressure_gradient_mult (Optional[float]): Correction Factor for the acceleration pressure
            gradient for pressure drop correlations (ACPGCR).
            con_type (Optional[str]): Connection type.
            hyd_method (optional[str | int]): Hydraulic method (METHOD)
            node_in (Optional[str]): Inflow node name.
            node_out (Optional[str]): Outflow node name.
            pvt_method (Optional[int]): PVT method (IPVT)
            water_method (Optional[int]): Water method (IWAT)
        """

        self.bat_method = bat_method
        self.elevation_profile = elevation_profile
        self.measured_depth_in = measured_depth_in
        self.measured_depth_out = measured_depth_out
        self.diameter = diameter
        self.inner_diameter = inner_diameter
        self.roughness = roughness
        self.heat_transfer_coeff = heat_transfer_coeff
        self.temperature_profile = temperature_profile
        self.length = length
        self.delta_depth = delta_depth
        self.connection_number = connection_number
        self.seawater_profile = seawater_profile
        self.rate_mult = rate_mult
        self.polymer = polymer
        self.dp_add = dp_add
        self.dt_add = dt_add
        self.temperature_in = temperature_in
        self.temperature_out = temperature_out
        self.tracer = tracer
        self.heat_conductivity = heat_conductivity
        self.insulation_thickness = insulation_thickness
        self.insulation_conductivity = insulation_conductivity
        self.gravity_pressure_gradient_mult = gravity_pressure_gradient_mult
        self.friction_pressure_gradient_mult = friction_pressure_gradient_mult
        self.acceleration_pressure_gradient_mult = acceleration_pressure_gradient_mult

        super().__init__(date_format=date_format, start_date=start_date, unit_system=unit_system, name=name, date=date,
                         properties_dict=properties_dict, node_in=node_in, node_out=node_out, con_type=con_type,
                         hyd_method=hyd_method, pvt_method=pvt_method, water_method=water_method)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        nexus_mapping = {
            'NAME': ('name', str),
            'NODEIN': ('node_in', str),
            'NODEOUT': ('node_out', str),
            'TYPE': ('con_type', str),
            'METHOD': ('hyd_method', str),
            'IPVT': ('pvt_method', int),
            'IWAT': ('water_method', int),
            'IBAT': ('bat_method', int),
            'ELEVPR': ('elevation_profile', str),
            'MDIN': ('measured_depth_in', float),
            'MDOUT': ('measured_depth_out', float),
            'DIAM': ('diameter', float),
            'INNERDIAM': ('inner_diameter', float),
            'ROUGHNESS': ('roughness', float),
            'HTC': ('heat_transfer_coeff', float),
            'TEMPPR': ('temperature_profile', str),
            'LENGTH': ('length', float),
            'DDEPTH': ('delta_depth', float),
            'NUMBER': ('connection_number', int),
            'SEAWPR': ('seawater_profile', str),
            'RATEMULT': ('rate_mult', float),
            'POLYMER': ('polymer', str),
            'DPADD': ('dp_add', float),
            'DTADD': ('dt_add', float),
            'TEMPIN': ('temperature_in', float),
            'TEMPOUT': ('temperature_out', float),
            'TRACERS': ('tracer', str),
            'HCOND': ('heat_conductivity', float),
            'INSTHN': ('insulation_thickness', float),
            'INSK': ('insulation_conductivity', float),
            'GRPGCR': ('gravity_pressure_gradient_mult', float),
            'FRPGCR': ('friction_pressure_gradient_mult', float),
            'ACPGCR': ('acceleration_pressure_gradient_mult', float),
        }
        return nexus_mapping



"""Abstract base class for holding the well connections data for holding all components of the production networks."""
from __future__ import annotations
from abc import ABC
from dataclasses import dataclass, field
from typing import Literal, Optional, TYPE_CHECKING, Sequence

from ResSimpy.DataModelBaseClasses.OperationsMixin import NetworkOperationsMixIn
from ResSimpy.DataModelBaseClasses.WellConnection import WellConnection
if TYPE_CHECKING:
    from ResSimpy.DataModelBaseClasses.Network import Network


@dataclass(kw_only=True)
class WellConnections(NetworkOperationsMixIn, ABC):
    """Abstract base class for holding the well connections data for holding all components of the production
    networks.
    """
    _well_connections: Sequence[WellConnection] = field(default_factory=list)

    def __init__(self, parent_network: Network) -> None:
        """Abstract base class for holding the well connections data for holding all components of the \
        production networks.

        Args:
            parent_network (Network): The parent network object that the well connections belong to.
        """
        super().__init__(parent_network)
        self.__parent_network: Network = parent_network
        self._well_connections = []

    @property
    def _network_element_name(self) -> Literal['well_connections']:
        return 'well_connections'

    def _add_to_memory(self, additional_list: Optional[list[WellConnection]]) -> None:
        """Extends the nodes object by a list of connections provided to it.

        Args:
            additional_list (Sequence[NexusWellConnection]): list of nexus connections to add to the nodes list.
        """
        if additional_list is None:
            return
        if not isinstance(additional_list, list):
            additional_list = list(additional_list)
        if not isinstance(self._well_connections, list):
            self._well_connections = list(self._well_connections)

        self._well_connections.extend(additional_list)



from __future__ import annotations
from abc import abstractmethod

from ResSimpy.Utils.invert_nexus_map import nexus_keyword_to_attribute_name
from typing import Protocol, Sequence, TYPE_CHECKING

from ResSimpy.Utils.to_dict_generic import to_dict

if TYPE_CHECKING:
    from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType


class SupportsKeywordMapping(Protocol):
    @staticmethod
    @abstractmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Returns a mapping of keywords.

        The values in dict can be of the following types:[str, tuple[str, type]].

        """
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def to_dict(self) -> DataObjectMixinDictType:
        """Writes a dictionary with the following types:[str, None | str | int | float | dict[int | float]]."""
        raise NotImplementedError("Implement this in the derived class")


def to_table_line(obj: SupportsKeywordMapping, headers: list[str]) -> str:
    """Takes a generic Nexus object and returns the attribute values as a string in the order of headers provided.

    Requires an implemented to_dict method and get_keyword_mapping() method.

    Args:
        obj (Any): a NexusObject with a to_dict and get_keyword_mapping method.
        headers (list[str]): list of header values in Nexus keyword format

    Returns:
        string of the values in the order of the headers provided.

    """
    nexus_mapping = obj.get_keyword_mapping()
    properties = obj.to_dict()

    constructed_values = []
    for header in headers:
        attribute_name = nexus_keyword_to_attribute_name(nexus_mapping, header)
        attribute_value = properties[attribute_name]
        if attribute_value is None:
            attribute_value = 'NA'
        elif attribute_value is True:
            attribute_value = 'YES'
        elif attribute_value is False:
            attribute_value = 'NO'
        constructed_values.append(attribute_value)
    return ' '.join([str(x) for x in constructed_values]) + '\n'


def get_column_headers_required(obj_list: Sequence[SupportsKeywordMapping]) -> list[str]:
    """Returns the column headers required to represent all non None attributes of the Nexus objects.

    Args:
        obj_list (list[SupportsKeywordMapping]): List of Nexus objects with a get_keyword_mapping method.

    Returns:
        list[str] representing the column headers in Nexus keyword format.
    """
    if not obj_list:
        return []
    headers: set[str] = set()
    for obj in obj_list:
        attr_as_dict = to_dict(obj, keys_in_nexus_style=True, include_nones=False, add_date=False, add_units=False)
        headers.update(attr_as_dict.keys())

    keyword_mapping = obj_list[0].get_keyword_mapping()
    headers_ordered_by_keyword_mapping = [x for x in keyword_mapping.keys() if x in headers]
    return headers_ordered_by_keyword_mapping



"""Abstract base class for network objects."""
from __future__ import annotations

import warnings
from abc import ABC
from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixin, DataObjectMixinDictType
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat


@dataclass(kw_only=True)
class NetworkObject(DataObjectMixin, ABC):
    """Abstract base class for an object that forms part of the surface network."""

    def __init__(self, properties_dict: Optional[DataObjectMixinDictType] = None,
                 date: Optional[str] = None, date_format: Optional[DateFormat] = None, start_date: Optional[str] = None,
                 unit_system: Optional[UnitSystem] = None, name: Optional[str] = None) -> None:
        """Initialises the NetworkObject class."""

        if properties_dict is None:
            properties_dict = {}

        if name is not None:
            obj_name = name
        else:
            dict_obj_name = properties_dict.get('name', None)
            if isinstance(dict_obj_name, str):
                obj_name = dict_obj_name
            else:
                obj_name = None

        if date is None:
            if 'date' not in properties_dict or not isinstance(properties_dict['date'], str):
                raise ValueError(f"No valid Date found for object with properties: {properties_dict}")
            date = properties_dict['date']

        if start_date is None:
            if 'start_date' in properties_dict:
                obj_start_date = properties_dict['start_date']
                start_date = obj_start_date if isinstance(obj_start_date, str) else None

        if date_format is None:
            if 'date_format' in properties_dict and isinstance(properties_dict['date_format'], DateFormat):
                date_format = properties_dict['date_format']
            else:
                raise ValueError("No date format supplied for object.")

        if unit_system is None:
            if 'unit_system' in properties_dict and isinstance(properties_dict['unit_system'], UnitSystem):
                unit_system = properties_dict['unit_system']
            else:
                # TODO: Update our unit tests with unit systems and upgrade the below to an error
                if 'Nexus' in self.__class__.__name__:
                    unit_system = UnitSystem.ENGLISH
                warnings.warn(f"No unit system supplied for object: {self.name}, assuming Simulator default of "
                              f"{unit_system}.")

        super().__init__(date_format=date_format, start_date=start_date, unit_system=unit_system, name=obj_name,
                         date=date)

        # Set attributes in the provided dict, excluding the ones already set.
        protected_attributes = ['date_format', 'start_date', 'unit_system', 'name']
        remaining_properties = [x for x in properties_dict.keys() if x not in protected_attributes]
        for key in remaining_properties:
            if properties_dict[key] is not None:
                self.__setattr__(key, properties_dict[key])



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Sequence

from ResSimpy.DataModelBaseClasses.SolverParameter import SolverParameter


@dataclass
class SolverParameters(ABC):
    """Abstract base class for storing solver parameter objects and handle loading of these parameters."""

    __solver_parameters: Sequence[SolverParameter]

    @abstractmethod
    def load(self) -> None:
        """Loads the solver parameters given to the simulator."""
        raise NotImplementedError("Implement this in the derived class.")

    @property
    @abstractmethod
    def solver_parameters(self) -> Sequence[SolverParameter]:
        """Returns an instance of Solver Parameter in a sequence (usually of type list)."""
        return self.__solver_parameters



from __future__ import annotations
from dataclasses import dataclass
from typing import TypeVar, TYPE_CHECKING

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixin
from ResSimpy.Enums.FluidTypeEnums import PvtType
from ResSimpy.Time.ISODateTime import ISODateTime
if TYPE_CHECKING:
    from ResSimpy import NexusSimulator
T = TypeVar('T', bound=DataObjectMixin)


@dataclass(kw_only=True)
class NexusModelFileGenerator:
    """A class to generate the contents of a NexusSimulator object as a set of model files."""
    __model: NexusSimulator
    __model_name: str

    def __init__(self, model: NexusSimulator, model_name: str) -> None:
        """Initializes the NexusModelFileGenerator."""
        self.__model = model
        self.__model_name = model_name

    @property
    def model(self) -> NexusSimulator:
        """The NexusSimulator model associated with this generator."""
        return self.__model

    def generate_base_model_file_contents(self) -> str:
        """Generates the base contents of the Nexus model file."""
        fcs_file_contents = self.model.model_files.to_string(dateformat=self.model.date_format,
                                                             run_units=self.model.run_units,
                                                             default_units=self.model.default_units)
        return fcs_file_contents

    def output_surface_section(self) -> str:
        """Outputs the surface section of the Nexus model file."""
        full_schedule = ''

        # add the pvt type and EOS properties:

        if self.model.pvt_type == PvtType.EOS and self.model.eos_details is not None:
            full_schedule += self.model.eos_details
        else:
            full_schedule += self.model.pvt_type.name
        full_schedule += '\n\n'

        if self.model.network is None:
            return full_schedule
        all_well_connections = self.model.network.well_connections.get_all()
        all_targets = self.model.network.targets.get_all()
        all_welllists = self.model.network.welllists.get_all()
        all_nodes = self.model.network.nodes.get_all()
        all_connections = self.model.network.connections.get_all()
        all_activations = self.model.network.activation_changes.get_all()

        network_objects = [all_well_connections, all_targets, all_welllists, all_nodes, all_connections,
                           all_activations]

        all_constraints = self.model.network.constraints.get_all()

        all_event_dates: set[ISODateTime] = set()

        all_event_dates.add(self.model.start_iso_date)
        for name, constraints in all_constraints.items():
            all_event_dates.update({x.iso_date for x in constraints})

        # get dates for all the items
        for net_obj in network_objects:
            all_event_dates.update({x.iso_date for x in net_obj})

        # Sort the dates
        ordered_all_event_dates = sorted(all_event_dates)

        # Write out all events for each date
        for date in ordered_all_event_dates:
            if date != self.model.start_iso_date:
                full_schedule += f"TIME {date.strftime_dateformat(self.model.date_format)}\n"

            well_connections_for_date = [x for x in all_well_connections if x.iso_date == date]

            if any(well_connections_for_date) and self.model.network.well_connections is not None:
                full_schedule += self.model.network.well_connections.to_string_for_date(date=date)
                full_schedule += '\n'

            welllists_for_date = [x for x in all_welllists if x.iso_date == date]
            if any(welllists_for_date) and self.model.network.welllists is not None:
                full_schedule += self.model.network.welllists.to_string_for_date(date=date)
                full_schedule += '\n'
            nodes_for_date = [x for x in all_nodes if x.iso_date == date]
            if any(nodes_for_date) and self.model.network.nodes is not None:
                full_schedule += self.model.network.nodes.to_string_for_date(date=date)
                full_schedule += '\n'
            connections_for_date = [x for x in all_connections if x.iso_date == date]
            if any(connections_for_date) and self.model.network.connections is not None:
                full_schedule += self.model.network.connections.to_string_for_date(date=date)
                full_schedule += '\n'

            constraints_for_date = {k: [x for x in v if x.iso_date == date] for k, v in all_constraints.items()}
            if any(x for x in constraints_for_date.values()) and self.model.network.constraints is not None:
                full_schedule += self.model.network.constraints.to_string_for_date(date=date)
                full_schedule += '\n'

            targets_for_date = [x for x in all_targets if x.iso_date == date]
            if any(targets_for_date) and self.model.network.targets is not None:
                full_schedule += self.model.network.targets.to_string_for_date(date=date)
                full_schedule += '\n'

            activations_for_date = [x for x in all_activations if x.iso_date == date]
            if any(activations_for_date) and self.model.network.activation_changes is not None:
                full_schedule += self.model.network.activation_changes.to_string_for_date(date=date)
                full_schedule += '\n'

        return full_schedule

    def output_wells_section(self) -> str:
        """Outputs the wells section of the Nexus model file."""
        if not self.model.wells._wells_loaded:
            return ''

        full_wells = ''

        all_well_completions = []
        all_well_mods = []
        for well in self.model.wells.get_all():
            all_well_completions.extend(well.completions)
            all_well_mods.extend(well.wellmods)

        well_mod_dates = {x.iso_date for x in all_well_mods}

        all_event_dates: set[ISODateTime] = self.model.wells.get_wells_iso_dates()
        all_event_dates.add(self.model.start_iso_date)
        all_event_dates.update(well_mod_dates)

        # Sort the dates
        ordered_all_event_dates = sorted(all_event_dates)

        # Write out all events for each date
        for date in ordered_all_event_dates:
            if date != self.model.start_iso_date:
                full_wells += f"TIME {date.strftime_dateformat(self.model.date_format)}\n"
            for well in self.model.wells.get_all():
                full_wells += well.to_string_for_date(date=date)

        return full_wells

    def output_options_section(self) -> str:
        """Outputs the options section of the Nexus model file."""
        options_file_content = ''
        if self.model.options is not None:
            options_file_content += self.model.options.to_string()
        # add the gridtoprocs to options:
        if self.model.sim_controls.grid_to_proc is not None:
            options_file_content += '\n'
            options_file_content += self.model.sim_controls.grid_to_proc.to_string()
        return options_file_content

    def output_runcontrol_section(self) -> str:
        """Outputs the run control section of the Nexus model file."""
        run_control_content = f'START {self.model.start_iso_date.strftime_dateformat(self.model.date_format)}\n\n'

        # collect all the dates from the solver controls
        all_dates: set[ISODateTime] = {self.model.start_iso_date}
        for solver_param in self.model.sim_controls.solver_parameters.get_all():
            all_dates.add(solver_param.iso_date)
        # collect all the dates from the reporting controls
        all_dates.update(self.model.reporting.get_all_reporting_dates())

        # add all the TIME cards:
        all_dates.update(set(self.model.sim_controls.times_iso_date))

        # Sort the dates
        ordered_all_dates = sorted(all_dates)

        for date in ordered_all_dates:
            if date != self.model.start_iso_date:
                run_control_content += f'\nTIME {date.strftime_dateformat(self.model.date_format)}\n'

            run_control_content += self.model.sim_controls.solver_parameters.to_string_for_date(date=date)
            run_control_content += self.model.reporting.to_string_for_date(date=date)

        return run_control_content



from dataclasses import dataclass
from typing import Optional, Mapping
from ResSimpy.Nexus.NexusKeywords.proc_keywords import PREDEFINED_VARS, CONDITIONAL_KEYWORDS, NEXUS_ALL_PROC_FUNCS


# set custom error message
class SetError(Exception):
    """A SetError message will appear when the user tries to wrongly set the attribute contents_breakdown."""


@dataclass(kw_only=True, repr=True)
class NexusProc:
    """Class that represents a single nexus procedure."""
    __date: Optional[str]
    __name: Optional[str]
    __priority: Optional[int]
    __contents: Optional[list[str]]
    __contents_breakdown: dict[str, int]

    def __init__(self, date: Optional[str] = None,
                 contents: Optional[list[str]] = None,
                 name: Optional[str] = None,
                 priority: Optional[int] = None) -> None:
        """Initialize the attributes of the class.

        Attributes:
                date: The date the procedure occured.
                name: The name of the procedure.
                priority: An integer describing the relative priority of the procedure.
                contents: The entire contents of the procedure (everything written between PROC and ENDPROC).
        contents_breakdown: A dictionary of counts for built-in nexus proc functions used in the procedure.

        Methods:
                reset_nexus_proc_function_counts(): Returns a dictionary of zero counts for the nexus proc functions.
        """

        self.__date = date
        self.__name = name
        self.__priority = priority
        self.__contents = contents
        self.__contents_breakdown = self.reset_nexus_proc_function_counts()

    @property
    def contents(self) -> Optional[list[str]]:
        """Returns the contents of the main body of the procedure."""
        return self.__contents

    @property
    def date(self) -> Optional[str]:
        """Returns the date that the procedure occurred."""
        return self.__date

    @property
    def priority(self) -> Optional[int]:
        """Returns the priority of the procedure."""
        return self.__priority

    @property
    def name(self) -> Optional[str]:
        """Returns the name of the procedure."""
        return self.__name

    @property
    def contents_breakdown(self) -> dict[str, int]:
        """Returns the dictionary counts of the proc functions."""
        return self.__contents_breakdown

    @contents_breakdown.setter
    def contents_breakdown(self, value: dict) -> None:
        # ensures this param is a dict with length equal to the Nexus proc functions
        if (isinstance(value, dict)) and (len(value) == len(self.reset_nexus_proc_function_counts())):
            self.__contents_breakdown = value
        else:
            raise SetError('This parameter is for internal use and should not be set by the user.\n'
                           'This parameter must be a dictionary where the keys are defined according to the 2022\n'
                           'Nexus Keyword Manual page 1013.')

    @staticmethod
    def get_keyword_mapping() -> Mapping[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords = {
            'NAME': ('name', str),
            'PRIORITY': ('priority', str)
        }
        return keywords

    @staticmethod
    def reset_nexus_proc_function_counts() -> dict[str, int]:
        """This function initializes and returns a dict of built-in Nexus proc functions.

        Please refer to page 1013 of the 2022 Nexus Keyword manual for a description of the functions.
            Returns: A dictionary of Nexus proc functions where the key is the specific function and the value is
        the count, initialized to zero.
        """

        predef_vars = dict(zip(PREDEFINED_VARS, [0] * len(PREDEFINED_VARS)))
        cond = dict(zip(CONDITIONAL_KEYWORDS, [0] * len(CONDITIONAL_KEYWORDS)))
        all_proc_funcs = dict(zip(NEXUS_ALL_PROC_FUNCS, [0] * len(NEXUS_ALL_PROC_FUNCS)))
        to_search = predef_vars | cond | all_proc_funcs

        return to_search



from __future__ import annotations
from dataclasses import dataclass, field
import os
from typing import Optional, MutableMapping

from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.DataModels.FcsFile import FcsNexusFile
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.DataModels.NexusHydraulicsMethod import NexusHydraulicsMethod
from ResSimpy.GenericContainerClasses.Hydraulics import Hydraulics
from ResSimpy.Utils.dynamic_method_manipulations import add_dynamic_method


@dataclass(kw_only=True)
class NexusHydraulicsMethods(Hydraulics):
    """Class for collection of Nexus hydraulics methods.

    Attributes:
        inputs (dict[int, NexusHydraulicsMethod]): Collection of Nexus hydraulics methods, as a dictionary
        files (dict[int, NexusFile]): Dictionary collection of hydraulics files, in Nexus fcs file.
    """

    __inputs: MutableMapping[int, NexusHydraulicsMethod]
    __files: dict[int, NexusFile]
    __properties_loaded: bool = False  # Used in lazy loading
    __model_unit_system: UnitSystem
    _model_files: Optional[FcsNexusFile] = field(default=None, repr=False, compare=False)

    def __init__(self, model_unit_system: UnitSystem,
                 inputs: Optional[MutableMapping[int, NexusHydraulicsMethod]] = None,
                 files: Optional[dict[int, NexusFile]] = None, assume_loaded: bool = False,
                 model_files: Optional[FcsNexusFile] = None) -> None:
        """Initialises the NexusHydraulicsMethods class.

        Args:
            model_unit_system (UnitSystem): Unit system used in the model.
            inputs (Optional[MutableMapping[int, NexusHydraulicsMethod]]): Collection of Nexus hydraulics methods.
            files (Optional[dict[int, NexusFile]]): Collection of hydraulics files, as defined in Nexus fcs file.
                Keyed by the method number.
            assume_loaded (bool): If True, assumes that the hydraulics methods are already loaded.
            model_files (Optional[FcsFile]): The FcsFile object that contains the model files.
        """
        if inputs:
            self.__inputs = inputs
        else:
            self.__inputs: MutableMapping[int, NexusHydraulicsMethod] = {}
        if files:
            self.__files = files
        else:
            self.__files = {}
        self.__model_unit_system = model_unit_system
        self.__properties_loaded = assume_loaded
        self._model_files = model_files
        super().__init__()

    def __repr__(self) -> str:
        """Pretty printing hydraulics methods."""
        if not self.__properties_loaded:
            self.load_hydraulics_methods()
        printable_str = ''
        for table_num in self.__inputs.keys():
            printable_str += '\n--------------------------------\n'
            printable_str += f'HYD method {table_num}\n'
            printable_str += '--------------------------------\n'
            printable_str += self.__inputs[table_num].__repr__()
            printable_str += '\n'

        return printable_str

    @property
    def inputs(self) -> MutableMapping[int, NexusHydraulicsMethod]:
        """Returns mapping of hydraulics method where keys are of type int.
        If properties are not loaded, it will load them first.
        """
        if not self.__properties_loaded:
            self.load_hydraulics_methods()
        return self.__inputs

    @property
    def files(self) -> dict[int, NexusFile]:
        """Returns a dictionary of Nexus files where keys are of type int."""
        return self.__files

    def load_hydraulics_methods(self) -> None:
        """Loads a collection of hydraulic method files defined by the Nexus fcs files."""
        # Read in hydraulics properties from Nexus hydraulics method files
        if self.__files is not None and len(self.__files) > 0:  # Check if hydraulics files exist
            for table_num in self.__files.keys():  # For each hydraulics property method
                hydraulics_file = self.__files[table_num]
                if hydraulics_file.location is None:
                    raise ValueError(f'Unable to find hydraulics file: {hydraulics_file}')
                if os.path.isfile(hydraulics_file.location):
                    # Create NexusHydraulicsMethod object
                    self.__inputs[table_num] = NexusHydraulicsMethod(file=hydraulics_file, input_number=table_num,
                                                                     model_unit_system=self.__model_unit_system)
                    self.__inputs[table_num].read_properties()  # Populate object with hydraulics props
        self.__properties_loaded = True

    @property
    def model_unit_system(self) -> UnitSystem:
        """Return the model unit system."""
        return self.__model_unit_system

    @property
    def summary(self) -> str:
        """Returns a string summary of 'NexusHydraulicsMethods' inputs in a dictionary."""
        hydraulics_summary = ''
        for key, value in self.inputs.items():
            file_loc = value.file.location if value.file else 'No file associated'
            hydraulics_summary += f'        {key}: {file_loc}\n'

        return hydraulics_summary

    def add_method(self, method: DynamicProperty, new_file_name: str, create_new_file: bool = False) -> None:
        """Adds a new hydraulics method to the collection.

        Args:
            method (DynamicProperty): The hydraulics method to add.
            new_file_name (str): The name of the file to save the method to.
            create_new_file (bool): Whether to create a new file for the method.
        """
        add_dynamic_method(dynamic_method_collection=self, method=method, new_file_name=new_file_name,
                           create_new_file=create_new_file)

    @property
    def keyword(self) -> str:
        """Returns the keyword for the hydraulics methods."""
        return 'HYD'

    def _method_type(self) -> type[NexusHydraulicsMethod]:
        """Returns the expected type of the dynamic property."""
        return NexusHydraulicsMethod



from __future__ import annotations
from abc import ABC
from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.DataModelBaseClasses.NetworkObject import NetworkObject
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Units.AttributeMappings.NetworkUnitMapping import NetworkUnits


@dataclass(repr=False)
class WellConnection(NetworkObject, ABC):
    bhdepth: Optional[float] = None
    datum_depth: Optional[float] = None
    x_pos: Optional[float] = None
    y_pos: Optional[float] = None
    length: Optional[float] = None
    temperature: Optional[float] = None
    diameter: Optional[float] = None
    roughness: Optional[float] = None
    inner_diameter: Optional[float] = None
    productivity_index: Optional[float] = None
    hyd_method: Optional[str] = None
    crossflow: Optional[str] = None
    crossshut: Optional[str] = None
    inj_mobility: Optional[str] = None
    polymer: Optional[str] = None
    stream: Optional[str] = None
    group: Optional[str] = None
    i: Optional[int] = None
    j: Optional[int] = None
    drainage_radius: Optional[float] = None
    gas_mobility: Optional[str] = None
    pvt_method: Optional[int] = None
    d_factor: Optional[float] = None
    on_time: Optional[float] = None
    well_index_mult: Optional[float] = None

    def __init__(self, properties_dict: Optional[DataObjectMixinDictType] = None,
                 date: Optional[str] = None, date_format: Optional[DateFormat] = None, start_date: Optional[str] = None,
                 unit_system: Optional[UnitSystem] = None, name: Optional[str] = None,
                 datum_depth: Optional[float] = None, crossflow: Optional[str] = None, crossshut: Optional[str] = None,
                 inj_mobility: Optional[str] = None, polymer: Optional[str] = None, stream: Optional[str] = None,
                 group: Optional[str] = None, i: Optional[int] = None, j: Optional[int] = None,
                 drainage_radius: Optional[float] = None, pvt_method: Optional[int] = None,
                 d_factor: Optional[float] = None, on_time: Optional[float] = None, bhdepth: Optional[float] = None,
                 x_pos: Optional[float] = None, y_pos: Optional[float] = None, length: Optional[float] = None,
                 temperature: Optional[float] = None, diameter: Optional[float] = None,
                 inner_diameter: Optional[float] = None, roughness: Optional[float] = None,
                 productivity_index: Optional[float] = None, hyd_method: Optional[str] = None,
                 well_index_mult: Optional[float] = None, gas_mobility: Optional[str] = None) -> None:
        """Initialises the WellConnection class.

        Args:
            properties_dict (dict): dict of the properties to set on the object.
            date (Optional[str]): The date of the object.
            date_format (Optional[DateFormat]): The date format that the object uses.
            start_date (Optional[str]): The start date of the model. Required if the object uses a decimal TIME.
            unit_system (Optional[UnitSystem]): The unit system of the object e.g. ENGLISH, METRIC.
            name (Optional[str]): The name of the object.
            crossshut (Optional[str]): Cross-shut identifier.
            crossflow (Optional[str]): Crossflow identifier.
            d_factor (Optional[float]): Non-Darcy D-factor (D).
            datum_depth (Optional[float]): Depth relative to a datum.
            drainage_radius (Optional[float]): The drainage radius.
            group (Optional[str]): The group that the Well Connection belongs to.
            i (Optional[int]): The location of the Well Connection in the i direction.
            j (Optional[int]): The location of the Well Connection in the j direction.
            inj_mobility (Optional[str]): Injection mobility identifier.
            on_time (Optional[float]): On-time duration.
            polymer (Optional[str]): Polymer identifier.
            pvt_method (Optional[int]): PVT method.
            stream (Optional[str]): Stream identifier.
            bhdepth (Optional[float]): Bottomhole depth (BHDEPTH).
            x_pos(Optional[float]): X-coordinate position (X).
            y_pos (Optional[float]): Y-coordinate position (Y).
            length (Optional[float]): Length of the well connection (LENGTH).
            temperature (Optional[float]): Temperature (TEMP).
            diameter (Optional[float]):  Diameter of the well connection (DIAM).
            inner_diameter (Optional[float]): Inner diameter of the well connection (INNERDIAM).
            roughness (Optional[float]): Pipe roughness (ROUGHNESS).
            productivity_index (Optional[float]): Productivity index (PI).
            hyd_method (Optional[str]): Hydraulic method (METHOD).
            well_index_mult (Optional[float]): Well index multiplier (WIMULT).
            gas_mobility (str): Gas mobility value (GASMOB).

        """

        self.bhdepth = bhdepth
        self.x_pos = x_pos
        self.y_pos = y_pos
        self.length = length
        self.temperature = temperature
        self.diameter = diameter
        self.inner_diameter = inner_diameter
        self.roughness = roughness
        self.productivity_index = productivity_index
        self.hyd_method = hyd_method
        self.well_index_mult = well_index_mult

        super().__init__(properties_dict=properties_dict, date=date, date_format=date_format, start_date=start_date,
                         unit_system=unit_system, name=name)

        self.datum_depth = datum_depth if datum_depth is not None else self.datum_depth
        self.crossflow = crossflow if crossflow is not None else self.crossflow
        self.crossshut = crossshut if crossshut is not None else self.crossshut
        self.inj_mobility = inj_mobility if inj_mobility is not None else self.inj_mobility
        self.polymer = polymer if polymer is not None else self.polymer
        self.stream = stream if stream is not None else self.stream
        self.group = group if group is not None else self.group
        self.i = i if i is not None else self.i
        self.j = j if j is not None else self.j
        self.drainage_radius = drainage_radius if drainage_radius is not None else self.drainage_radius
        self.pvt_method = pvt_method if pvt_method is not None else self.pvt_method
        self.d_factor = d_factor if d_factor is not None else self.d_factor
        self.on_time = on_time if on_time is not None else self.on_time
        self.well_index_mult = well_index_mult if well_index_mult is not None else self.well_index_mult
        self.gas_mobility = gas_mobility if gas_mobility is not None else self.gas_mobility

    @property
    def units(self) -> NetworkUnits:
        """Returns the attribute to unit map for the WellConnection."""
        return NetworkUnits(self.unit_system)



from dataclasses import dataclass

from ResSimpy.DataModelBaseClasses.Over import Over


@dataclass
class NexusOver(Over):
    """Used to represent a line from an OVER table in a Nexus file.

    Attributes:
        array (str): The arrays to be modified by the over.
        i1 (int): The start of the i range.
        i2 (int): The end of the i range.
        j1 (int): The start of the j range.
        j2 (int): The end of the j range.
        k1 (int): The start of the k range.
        k2 (int): The end of the k range.
        operator (str): The operator to be assigned to the range.
        value (float): The value to be assigned to the range.
        threshold (float): The value for which GE or LE is used.
    """

    def to_string(self) -> str:
        """Converts the over to a string for writing to a Nexus file."""
        return_str = 'OVER'
        return_str += f" {self.array}"
        return_str += '\n'
        if self.grid != 'ROOT':
            return_str += f"GRID {self.grid}\n"
        if self.fault_name:
            return_str += f"FNAME {self.fault_name}\n"
        return_str += f"{self.i1} {self.i2} {self.j1} {self.j2} {self.k1} {self.k2} {self.operator}{self.value}\n"
        return return_str



"""The base class for all Well Completions."""
from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, Sequence

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixin
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.NexusEnums import DateFormatEnum
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Units.AttributeMappings.CompletionUnitMapping import CompletionUnits


@dataclass(kw_only=True, repr=False)
class Completion(DataObjectMixin, ABC):
    """A class representing well completions.

    IMPORTANT: if modifying this class, make sure to update the relevant tests in test_load_wells, as well as updating
    the constructor calls in the derived classes.

    Args:
    ----
        date (str): The starting date of the completion. Applies until changed.
        i (Optional[int]): The structured grid cell location in the x direction. 'IW' in Nexus
        j (Optional[int]): The structured grid cell location in the y direction. 'JW' in Nexus
        k (Optional[int]): The structured grid cell location in the z direction. 'L' in Nexus
        skin (Optional[float]): The skin value for the completion. 'SKIN' in Nexus
        depth (Optional[float]): The depth of the completion. 'DEPTH' in Nexus
        well_radius (Optional[float]): The well radius. 'RADW' in Nexus
        x (Optional[float]): The x location of the well in distance units/coordinates. 'X' in Nexus
        y (Optional[float]): The y location of the well in distance units/coordinates. 'Y' in Nexus
        angle_a (Optional[float]): the angle relative to the local I axis. 'ANGLA' in Nexus.
        angle_v (Optional[float]): the angle relative to the true vertical axis (global Z axis). 'ANGLV' in Nexus
        grid (Optional[str]): the grid name to which the completion data applies. 'GRID' in Nexus
        depth_to_top (Optional[float]): subsea depth to the top of a completion interval. 'DTOP' in Nexus
        depth_to_bot (Optional[float]): subsea depth to the bottom of the completion interval. 'DBOT' in Nexus
        perm_thickness_ovr (Optional[float]): permeability thickness override value to use for the completion interval.\
            'KH' in Nexus.
        dfactor (Optional[float]): non-darcy factor to use for rate dependent skin calculations. 'D' in Nexus
        rel_perm_method (Optional[int]): rel perm method to use for the completion. 'IRELPM' in Nexus
        status (Optional[str]): the status of the layer, can be 'ON' or 'OFF'
        peaceman_well_block_radius (Optional[float]): The pressure equivalent radius of the grid block
    """

    _i: Optional[int] = None
    _j: Optional[int] = None
    _k: Optional[int] = None
    _skin: Optional[float] = None
    _depth: Optional[float] = None
    _well_radius: Optional[float] = None
    _x: Optional[float] = None
    _y: Optional[float] = None
    _angle_a: Optional[float] = None
    _angle_v: Optional[float] = None
    _grid: Optional[str] = None
    _depth_to_top: Optional[float] = None
    _depth_to_bottom: Optional[float] = None
    _perm_thickness_ovr: Optional[float] = None
    _dfactor: Optional[float] = None
    _rel_perm_method: Optional[int] = None
    _status: Optional[str] = None
    _peaceman_well_block_radius: Optional[float] = None
    _perm_thickness_mult: Optional[float] = None

    def __init__(self, date: str, i: Optional[int] = None, j: Optional[int] = None, k: Optional[int] = None,
                 skin: Optional[float] = None, depth: Optional[float] = None, well_radius: Optional[float] = None,
                 x: Optional[float] = None, y: Optional[float] = None, angle_a: Optional[float] = None,
                 angle_v: Optional[float] = None, grid: Optional[str] = None, depth_to_top: Optional[float] = None,
                 depth_to_bottom: Optional[float] = None, perm_thickness_ovr: Optional[float] = None,
                 dfactor: Optional[float] = None, rel_perm_method: Optional[int] = None,
                 perm_thickness_mult: Optional[float] = None, status: Optional[str] = None,
                 date_format: Optional[DateFormatEnum.DateFormat] = None,
                 peaceman_well_block_radius: Optional[float] = None, start_date: Optional[str] = None,
                 unit_system: Optional[UnitSystem] = None) -> None:
        """Initialises the Completion class.

        Args:
            date: (str): The starting date of the completion.
            i: (Optional[int]): The structured grid cell location in the x direction.
            j: (Optional[int]): The structured grid cell location in the y direction.
            k: Optional[int]: The structured grid cell location in the z direction.
            skin: Optional[float]: The skin value for the completion.
            depth: Optional[float]: The depth of the completion.
            well_radius: Optional[float]: The well radius.
            x: Optional[float]: The x location of the well in distance units/coordinates.
            y: Optional[float]: The y location of the well in distance units/coordinates.
            angle_a: Optional[float]: the angle relative to the local I axis.
            angle_v: Optional[float]: the angle relative to the true vertical axis (global Z axis).
            grid: Optional[str]: the grid name to which the completion data applies.
            depth_to_top: Optional[float]: depth to the top of a completion interval.
            depth_to_bottom: Optional[float]: depth to the bottom of the completion interval.
            perm_thickness_ovr: Optional[float]: permeability thickness override value to use for the
            completion interval.
            dfactor: Optional[float]: non-darcy factor to use for rate dependent skin calculations.
            rel_perm_method: Optional[int]: rel perm method to use for the completion.
            perm_thickness_mult: Optional[float]: permeability thickness multiplier to use for the completion interval.
            status: Optional[str]: the status of the layer, can be 'ON' or 'OFF'
            date_format: Optional[DateFormatEnum.DateFormat]: The date format to use for the date as an enum.
            peaceman_well_block_radius: Optional[float]: The pressure equivalent radius of the grid block
            start_date: Optional[str]: The start date of the simulation.
            unit_system: Optional[UnitSystem]: The unit system to use for the completion.
        """
        super().__init__(date=date, date_format=date_format, start_date=start_date, unit_system=unit_system)
        self._well_radius = well_radius
        self._i = i
        self._j = j
        self._k = k
        self._skin = skin
        self._depth = depth
        self._x = x
        self._y = y
        self._angle_a = angle_a
        self._angle_v = angle_v
        self._grid = grid
        self._depth_to_top = depth_to_top
        self._depth_to_bottom = depth_to_bottom
        self._perm_thickness_ovr = perm_thickness_ovr
        self._dfactor = dfactor
        self._rel_perm_method = rel_perm_method
        self._perm_thickness_mult = perm_thickness_mult
        self._status = status
        self.__unit_system = unit_system
        self._peaceman_well_block_radius = peaceman_well_block_radius

    @property
    def well_radius(self) -> float | None:
        """Returns well radius as float."""
        return self._well_radius

    @property
    def i(self) -> int | None:
        """Returns grid cell location in the x direction as an int."""
        return self._i

    @property
    def j(self) -> int | None:
        """Returns grid cell location in the y direction as an int."""
        return self._j

    @property
    def k(self) -> int | None:
        """Returns grid cell location in the z direction as an int."""
        return self._k

    @property
    def skin(self) -> float | None:
        """Returns skin value of the completion as a float."""
        return self._skin

    @property
    def depth(self) -> float | None:
        """Returns the depth of the completion as a float."""
        return self._depth

    @property
    def x(self) -> float | None:
        """Returns the location of the completion in the x direction."""
        return self._x

    @property
    def y(self) -> float | None:
        """Returns the location of the completion in the y direction."""
        return self._y

    @property
    def angle_a(self) -> float | None:
        """Returns the angle relative to the local I axis."""
        return self._angle_a

    @property
    def angle_v(self) -> float | None:
        """Returns the angle relative to the true vertical axis (global Z axis)."""
        return self._angle_v

    @property
    def grid(self) -> str | None:
        """Returns the grid name to which the completion data applies."""
        return self._grid

    @property
    def depth_to_top(self) -> float | None:
        """Returns the depth to the top of the completion interval as a float."""
        return self._depth_to_top

    @property
    def depth_to_bottom(self) -> float | None:
        """Returns the depth to the bottom as a float."""
        return self._depth_to_bottom

    @property
    def perm_thickness_ovr(self) -> float | None:
        """Returns premeability thickness override value as a float to use for the completion interval."""
        return self._perm_thickness_ovr

    @property
    def dfactor(self) -> float | None:
        """Returns the non-darcy factor to use for rate dependent skin calculations as a float."""
        return self._dfactor

    @property
    def rel_perm_method(self) -> int | None:
        """Writes the relative permeability used for completion as an int."""
        return self._rel_perm_method

    @property
    def status(self) -> str | None:
        """Returns the status of the layer as a string whether it is 'ON' or 'OF'."""
        return self._status

    @property
    def peaceman_well_block_radius(self) -> Optional[float]:
        """Writes the pressure equivalent radius of the grid block."""
        return self._peaceman_well_block_radius

    @property
    def perm_thickness_mult(self) -> Optional[float]:
        """Returns the permeability thickness multiplier to use for the completion interval."""
        return self._perm_thickness_mult

    @property
    def date_format(self) -> Optional[DateFormat]:
        """Returns date format."""
        return self._date_format

    @property
    def unit_system(self) -> UnitSystem | None:
        """Returns unit system for the completion."""
        return self.__unit_system

    @property
    def units(self) -> CompletionUnits:
        """Writes an instance of the Completion Units class initialized in unit system of the object."""
        return CompletionUnits(self.unit_system)

    @property
    def completion_is_shutin(self) -> bool:
        """Returns a boolean indicating whether the provided completion is categorized as a shut-in or not."""
        return not self.completion_is_perforation

    @property
    def perforations(self) -> Sequence[Completion]:
        """Returns a list of all the perforations for the well."""
        raise NotImplementedError("This method has not been implemented for this simulator yet")

    @property
    def first_perforation(self) -> Optional[Completion]:
        """Returns the first perforation for the well."""
        raise NotImplementedError("This method has not been implemented for this simulator yet")

    @property
    def shutins(self) -> Sequence[Completion]:
        """Returns a list of all the perforations for the well."""
        raise NotImplementedError("This method has not been implemented for this simulator yet")

    @property
    def last_shutin(self) -> Optional[Completion]:
        """Returns the first perforation for the well."""
        raise NotImplementedError("This method has not been implemented for this simulator yet")

    @property
    @abstractmethod
    def completion_is_perforation(self) -> bool:
        """Determines if the supplied completion is a perforation or not."""
        raise NotImplementedError("Implement this in the derived class")



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Literal, Sequence, TYPE_CHECKING, Optional

from ResSimpy.DataModelBaseClasses.NetworkList import NetworkList
from ResSimpy.Time.ISODateTime import ISODateTime

if TYPE_CHECKING:
    from ResSimpy.DataModelBaseClasses.Network import Network


@dataclass(kw_only=True)
class NetworkLists(ABC):
    """Base class for representing different list types for the model."""
    _lists: Sequence[NetworkList]
    __parent_network: Network

    def __init__(self, parent_network: Network) -> None:
        """Initialises the NetworkList ABC."""
        self.__parent_network: Network = parent_network

    @property
    @abstractmethod
    def _network_element_name(self) -> Literal['conlists', 'welllists', 'nodelists']:
        raise NotImplementedError('This method must be implemented in the derived class.')

    @property
    def lists(self) -> Sequence[NetworkList]:
        """Returns all NetworkList instances."""
        return self._lists

    @property
    def unique_names(self) -> list[str]:
        """Returns all NetworkList names."""
        return list({x.name for x in self.lists if x.name is not None})

    @staticmethod
    @abstractmethod
    def table_header() -> Literal['WELLLIST', 'CONLIST', 'NODELIST']:
        """Start of the table."""
        raise NotImplementedError('This method must be implemented in the derived class.')

    @staticmethod
    @abstractmethod
    def table_footer() -> Literal['ENDWELLLIST', 'ENDCONLIST', 'ENDNODELIST']:
        """End of the table."""
        raise NotImplementedError('This method must be implemented in the derived class.')

    def get_all_by_name(self, list_name: str) -> Sequence[NetworkList]:
        """Returns a list of NetworkLists which match the provided name loaded from the simulator.

        Args:
        ----
            list_name (str): name of the requested ConList

        Returns:
        -------
            list[NexusConList]: the requested list of filtered lists with the same name
        """
        self.__parent_network.get_load_status()
        return [x for x in self._lists if x.name == list_name]

    @abstractmethod
    def _add_to_memory(self, additional_list: Optional[Sequence[NetworkList]]) -> None:
        """Adds additional lists to the current list."""
        raise NotImplementedError('This method must be implemented in the derived class.')

    def to_string_for_date(self, date: ISODateTime) -> str:
        """Returns a string representation of the lists for a specific date."""
        lists_for_date = [x for x in self._lists if x.iso_date == date]
        if not lists_for_date:
            return ''
        printable_string = ''
        for list_item in lists_for_date:
            printable_string += f'{self.table_header()} {list_item.name}\n'
            # remove all the previous wells and reinitialise the list
            printable_string += 'CLEAR\nADD\n'
            printable_string += ' '.join(list_item.elements_in_the_list) + '\n'
            printable_string += self.table_footer() + '\n'
        return printable_string



"""Enum for defining types of outputs."""
from enum import Enum


class OutputType(Enum):
    """Enum for defining types of outputs."""

    UNDEFINED = 0
    NONE = 1
    ARRAY = 2
    SPREADSHEET = 4
    MAP = 5



from __future__ import annotations
from abc import ABC
from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.DataModelBaseClasses.NetworkObject import NetworkObject
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Units.AttributeMappings.NetworkUnitMapping import NetworkUnits


@dataclass(repr=False)
class Wellhead(NetworkObject, ABC):
    well: Optional[str] = None
    wellhead_type: Optional[str] = None
    depth: Optional[float] = None
    x_pos: Optional[float] = None
    y_pos: Optional[float] = None

    def __init__(self, properties_dict: Optional[DataObjectMixinDictType] = None,
                 date: Optional[str] = None, date_format: Optional[DateFormat] = None, start_date: Optional[str] = None,
                 unit_system: Optional[UnitSystem] = None, name: Optional[str] = None, well: Optional[str] = None,
                 wellhead_type: Optional[str] = None, depth: Optional[float] = None, x_pos: Optional[float] = None,
                 y_pos: Optional[float] = None) -> None:
        """Initialises the Wellhead class.

        Args:
            properties_dict (dict): dict of the properties to set on the object.
            date (Optional[str]): The date of the object.
            date_format (Optional[DateFormat]): The date format that the object uses.
            start_date (Optional[str]): The start date of the model. Required if the object uses a decimal TIME.
            unit_system (Optional[UnitSystem]): The unit system of the object e.g. ENGLISH, METRIC.
            name (Optional[str]): The name of the object.
            well (Optional[str]): Associates the wellhead to the well.
            wellhead_type (Optional[str]): The type of well.
            depth (Optional[float]): The depth of the wellhead
            x_pos (Optional[float]): The x-coordinate of the wellhead (x).
            y_pos (Optional[float]): The y-coordinate of the wellhead (y).
        """

        self.well = well
        self.wellhead_type = wellhead_type
        self.depth = depth
        self.x_pos = x_pos
        self.y_pos = y_pos

        super().__init__(properties_dict=properties_dict, date=date, date_format=date_format, start_date=start_date,
                         unit_system=unit_system, name=name)

    @property
    def units(self) -> NetworkUnits:
        """Returns the attribute to unit map for the WellConnection."""
        return NetworkUnits(self.unit_system)



"""Creates a repr for data objects that removes attributes that are None from the representation."""
from typing import Any, Optional


def generic_repr(input_class: Any, exclude_attributes: Optional[list[str]] = None) -> str:
    """Creates a prettier object representation while removing attributes that are None from that representation.

    Args:
    ----
        input_class (Any): a class with attributes to summarise
        exclude_attributes (list[str]): a list of attributes to exclude. Defaults to None.

    Returns:
    -------
        (str): Pretty representation of the string.
    """
    filtered_attrs = {k: v for k, v in vars(input_class).items() if v is not None}

    # Remove the leading underscores from the repr.
    sanitised_attrs = {}
    for key in filtered_attrs.keys():
        if exclude_attributes is not None and key in exclude_attributes:
            continue
        elif key == '_DataObjectMixin__id':
            sanitised_key = 'ID'
        elif key == '_DataObjectMixin__iso_date':
            sanitised_key = 'ISO_Date'
        elif key == '_DataObjectMixin__date':
            sanitised_key = 'Date'
        elif key[0] == '_':
            sanitised_key = key[1:] if key[1] != '_' else key[2:]
        else:
            sanitised_key = key

        sanitised_attrs[sanitised_key] = filtered_attrs[key] if sanitised_key != 'ISO_Date' \
            else str(filtered_attrs[key])

    attrs = ', '.join(f"{k}={v!r}" for k, v in sanitised_attrs.items())
    return f"{input_class.__class__.__name__}({attrs})"


def generic_str(input_class: Any) -> str:
    """Creates a string representation of the object while removing None and ids from the string.

    Args:
        input_class (Any): a class with attributes to summarise

    Returns:
        (str): String representation of the object.
    """
    filtered_attrs = {k: v for k, v in vars(input_class).items() if v is not None}
    id_keys = [id_attr for id_attr in filtered_attrs if id_attr.endswith('__id')]
    for id_attr in id_keys:
        del filtered_attrs[id_attr]

    attrs = ', '.join(f"{k}={v!r}" for k, v in filtered_attrs.items())
    return f"{input_class.__class__.__name__}({attrs})"



"""Class to hold data input for a Nexus Separator method."""
from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Union
import numpy as np
import pandas as pd
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Utils.factory_methods import get_empty_dict_union
import ResSimpy.Nexus.nexus_file_operations as nfo
from ResSimpy.Nexus.NexusKeywords.separator_keywords import SEPARATOR_KEYS_INT, SEPARATOR_KEYS_FLOAT
from ResSimpy.Nexus.NexusKeywords.separator_keywords import SEPARATOR_KEYWORDS
from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.Units.AttributeMappings.DynamicPropertyUnitMapping import SeparatorUnits
from ResSimpy.Enums.UnitsEnum import UnitSystem, SUnits, TemperatureUnits
from ResSimpy.Enums.FluidTypeEnums import SeparatorType


@dataclass(kw_only=True, repr=False)  # Doesn't need to write an _init_, _eq_ methods, etc.
class NexusSeparatorMethod(DynamicProperty):
    """Class to hold data input for a Nexus Separator method.

    Attributes:
        file (NexusFile): Nexus Separator file object
        input_number (int): Separator method number in Nexus fcs file
        separator_type (Optional[str]): Type of separator method, e.g., BLACKOIL, GASPLANT or EOS. Defaults to None
        properties (dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                    dict[str, Union[float, pd.DataFrame]]]]):
            Dictionary holding all properties for a specific separator method. Defaults to empty dictionary.
    """

    file: NexusFile
    separator_type: Optional[SeparatorType] = None
    properties: dict[str, Union[str, int, float, Enum, list[str], np.ndarray,
                     pd.DataFrame, dict[str, Union[float, pd.DataFrame]]]] \
        = field(default_factory=get_empty_dict_union)
    unit_system: UnitSystem
    __hash__ = DynamicProperty.__hash__

    def __init__(self, file: NexusFile, input_number: int, model_unit_system: UnitSystem,
                 separator_type: Optional[SeparatorType] = None,
                 properties: Optional[dict[str, Union[str, int, float, Enum, list[str], np.ndarray, pd.DataFrame,
                                      dict[str, Union[float, pd.DataFrame]]]]] = None) -> None:
        """Initialises the NexusSeparatorMethod class.

        Args:
            file (NexusFile): NexusFile object associated with the separator method
            input_number (int): method number for the separator method
            model_unit_system (UnitSystem): unit system from the model
            separator_type (Optional[SeparatorType]): type of separator method, e.g., BLACKOIL, GASPLANT or EOS

            properties: dictionary of properties for the separator method.
        """
        if separator_type is not None:
            self.separator_type = separator_type
        if properties is not None:
            self.properties = properties
        else:
            self.properties = {}
        self.unit_system = model_unit_system
        super().__init__(input_number=input_number, file=file)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords: dict[str, tuple[str, type]] = {
            'TEMP': ('temperature', float),
            'PRES': ('pressure', float),
            'TEMP_STD': ('standard_temperature', float),
            'PRES_STD': ('standard_pressure', float),
        }
        return keywords

    @property
    def units(self) -> SeparatorUnits:
        """Returns the attribute to unit map for the separator method."""
        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']
        return SeparatorUnits(unit_system=self.unit_system)

    def to_string(self) -> str:
        """Create string with separator data in Nexus file format."""
        printable_str = ''
        sep_dict = self.properties
        for key, value in sep_dict.items():
            if key == 'DESC' and isinstance(value, list):
                for desc_line in value:
                    printable_str += 'DESC ' + desc_line + '\n'
            elif key == 'SEPARATOR_TABLE' and isinstance(value, pd.DataFrame):
                if self.separator_type == SeparatorType.GASPLANT:
                    printable_str += 'RECFAC_TABLE\n'
                printable_str += value.to_string(na_rep='', index=False) + '\n'
                if self.separator_type == SeparatorType.GASPLANT:
                    printable_str += 'ENDRECFAC_TABLE\n'
                printable_str += '\n'
            elif isinstance(value, Enum):
                if isinstance(value, UnitSystem) or isinstance(value, TemperatureUnits):
                    printable_str += f'{value.value}\n'
                elif isinstance(value, SUnits):
                    printable_str += f'SUNITS {value.value}\n'
            elif key == 'KEYCPTLIST' and isinstance(value, list):
                printable_str += f"{key} {' '.join(value)}\n"
            elif value == '':
                printable_str += f'{key}\n'
            else:
                printable_str += f'{key} {value}\n'
        return printable_str

    def read_properties(self) -> None:
        """Read Nexus Separator file contents and populate NexusSeparatorMethod object."""
        file_as_list = self.file.get_flat_list_str_file

        # Check for common input data
        nfo.check_for_and_populate_common_input_data(file_as_list, self.properties)

        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']

        # Initialize flags and containers used to record properties as we iterate through separator file contents
        # List to record start and ending indices for tables
        sep_table_indices: list[int] = [0, len(file_as_list)]
        # Flag to tell when to start reading a table
        start_reading_table: bool = False

        line_indx = 0
        for line in file_as_list:

            # Determine separator type, i.e., EOS multistage, gas plant data or black oil
            if nfo.check_token('TEMP', line):  # EOS multistage separator table
                self.separator_type = SeparatorType.EOS
                sep_table_indices[0] = line_indx  # Specify EOS multistage separator table starting index
                start_reading_table = True
                line_indx += 1
                continue
            elif nfo.check_token('KEYCPTLIST', line):  # Gas plant data
                self.separator_type = SeparatorType.GASPLANT
                elems = line.split('!')[0].split()
                cpt_index = elems.index('KEYCPTLIST')
                if 'KEYCPTLIST' not in self.properties.keys():
                    self.properties['KEYCPTLIST'] = elems[cpt_index + 1:]
                line_indx += 1
                continue
            elif nfo.check_token('BOSEP', line):  # Black oil separator table
                self.separator_type = SeparatorType.BLACKOIL
                self.properties['BOSEP'] = ''
                line_indx += 1
                continue

            # Find SEPARATOR key-value pairs, such as WATERMETHOD 1 or PRES_STD 14.7
            if [i for i in line.split() if i in SEPARATOR_KEYS_FLOAT]:
                for key in SEPARATOR_KEYS_FLOAT:
                    if nfo.check_token(key, line):
                        self.properties[key] = float(nfo.get_expected_token_value(key, line, file_as_list))
            if [i for i in line.split() if i in SEPARATOR_KEYS_INT]:
                for key in SEPARATOR_KEYS_INT:
                    if nfo.check_token(key, line):
                        self.properties[key] = int(nfo.get_expected_token_value(key, line, file_as_list))

            # Find starting index for black oil separator table
            if self.separator_type == SeparatorType.BLACKOIL:
                if nfo.check_token('KVOIL', line):
                    sep_table_indices[0] = line_indx
                    start_reading_table = True
                    line_indx += 1
                    continue

            # Find ending index for EOS multistage and black oil separator tables
            if start_reading_table:
                if self.separator_type in [SeparatorType.EOS, SeparatorType.BLACKOIL]:
                    for keyword in SEPARATOR_KEYWORDS:
                        if nfo.check_token(keyword, line):
                            sep_table_indices[1] = line_indx
                            start_reading_table = False
                            break

            # Find starting and ending indices for gas plant separator table
            if self.separator_type == SeparatorType.GASPLANT:
                if nfo.check_token('RECFAC_TABLE', line):
                    sep_table_indices[0] = line_indx + 1
                    start_reading_table = True
                if nfo.check_token('ENDRECFAC_TABLE', line):
                    sep_table_indices[1] = line_indx
                    start_reading_table = False

            line_indx += 1

        # Read in separator table
        if self.separator_type is not None:
            self.properties['SEPARATOR_TABLE'] = nfo.read_table_to_df(file_as_list[sep_table_indices[0]:
                                                                                   sep_table_indices[1]])



"""Base class representing a data object in ResSimpy."""
from uuid import uuid4, UUID
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Optional

from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Time.ISODateTime import ISODateTime
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Units.AttributeMappings.BaseUnitMapping import BaseUnitMapping
from ResSimpy.Utils import to_dict_generic
from ResSimpy.Utils.generic_repr import generic_repr, generic_str
from ResSimpy.Utils.obj_to_table_string import to_table_line

# generic type for dict
DataObjectMixinDictType = dict[str, None | str | int | float | dict[int, float]]


@dataclass(kw_only=True)
class DataObjectMixin(ABC):
    """Base class representing a data object in ResSimpy."""
    __id: UUID = field(default_factory=lambda: uuid4(), compare=False)
    __iso_date: ISODateTime = field(init=False, repr=True)
    _date_format: Optional[DateFormat] = None
    __date: Optional[str] = None
    _start_date: Optional[str] = None
    _unit_system: Optional[UnitSystem] = None
    __name: Optional[str] = None

    def __init__(self, date: Optional[str] = None, date_format: Optional[DateFormat] = None,
                 start_date: Optional[str] = None, unit_system: Optional[UnitSystem] = None,
                 name: Optional[str] = None) -> None:
        """Initialises the DataObjectMixin Class. First '_' parameter is a dummy parameter for type compatibility.

        Args:
            name (Optional[str]): The name of the object.
            date (Optional[str]): The date of the object.
            date_format (Optional[DateFormat]): The date format of the object.
            start_date (Optional[str]): The start date of the model (required if the date is in numerical format).
            unit_system (Optional[UnitSystem]): The unit system of the object.
        """
        self.__id = uuid4()
        self._date_format = date_format
        self._start_date = start_date
        self._unit_system = unit_system
        self.date = date
        self.__name = name

    def __repr__(self) -> str:
        return generic_repr(self)

    def __str__(self) -> str:
        return generic_str(self)

    @property
    def iso_date(self) -> ISODateTime:
        """The date of the object in ISO format."""
        if self.__iso_date is None:
            self.set_iso_date()
        return self.__iso_date

    @property
    def date(self) -> Optional[str]:
        """The date of the object as written in the source file."""
        return self.__date

    @date.setter
    def date(self, value: None | str) -> None:
        self.__date = value
        if value is not None:
            self.set_iso_date()

    @property
    def start_date(self) -> Optional[str]:
        """The start date of the model (required for Nexus objects using a decimal date format)."""
        return self._start_date

    @property
    def date_format(self) -> Optional[DateFormat]:
        """The date format of the date on the object."""
        return self._date_format

    @property
    def unit_system(self) -> Optional[UnitSystem]:
        """The unit system being used in the object."""
        return self._unit_system

    @property
    def name(self) -> Optional[str]:
        """The name of the object in the simulator."""
        return self.__name

    def set_iso_date(self) -> None:
        """Updates the ISO Date property."""
        if self.date is None:
            raise ValueError("Cannot set ISO Date without a date.")

        self.__iso_date = ISODateTime.convert_to_iso(self.date, self.date_format, self.start_date)

    def to_dict(self, keys_in_keyword_style: bool = False, add_date: bool = True, add_units: bool = True,
                add_iso_date: bool = False, include_nones: bool = True,
                units_as_string: bool = True) -> DataObjectMixinDictType:
        """Returns a dictionary of the attributes of the object.

        Args:
            keys_in_keyword_style (bool): if True returns the key values as simulator keywords, otherwise returns the \
                attribute name as stored by ressimpy.
            add_date (bool): if True adds the date to the dictionary
            add_iso_date (bool): adds an iso date attribute if it exists.
            add_units (bool): if True adds the units to the dictionary
            include_nones (bool): if True includes None values from the object in the dictionary.
            units_as_string (bool): If True, converts the object's units to a string value rather than an enum.

        Returns:
            a dictionary keyed by attributes and values as the value of the attribute
        """
        result_dict = to_dict_generic.to_dict(self, keys_in_keyword_style, add_date=add_date, add_units=add_units,
                                              add_iso_date=add_iso_date,
                                              include_nones=include_nones, units_as_string=units_as_string)
        return result_dict

    def to_table_line(self, headers: list[str]) -> str:
        """Takes a generic Nexus object and returns the attribute values as a string in the order of headers provided.

        Requires an implemented to_dict method and get_keyword_mapping() method.

        Args:
            headers (list[str]): list of header values in keyword format

        Returns:
            string of the values in the order of the headers provided.

        """
        return to_table_line(self, headers)

    @property
    def id(self) -> UUID:
        """Unique identifier for each object."""
        return self.__id

    @staticmethod
    @abstractmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of keywords to attribute definitions."""
        raise NotImplementedError("Implement this in the derived class")

    @property
    @abstractmethod
    def units(self) -> BaseUnitMapping:
        """Returns the attribute to unit map for the data object."""
        raise NotImplementedError("Implement this in the derived class")

    def get_unit_for_attribute(self, attribute_name: str, uppercase: bool = False) -> str:
        """Returns the unit variable for the given unit system.

        Args:
            attribute_name (str): name of the attribute to get the unit for
            unit_system (UnitSystem): unit system to get the unit for
            uppercase (bool): if True returns the unit in uppercase
        """
        unit_dimension = self.units.attribute_map.get(attribute_name, None)
        if unit_dimension is None:
            raise AttributeError(f'Attribute {attribute_name} not recognised and does not have a unit definition')

        if self.unit_system is None:
            raise AttributeError("Cannot find unit without a unit system.")

        unit = unit_dimension.unit_system_enum_to_variable(unit_system=self.unit_system)
        if uppercase:
            unit = unit.upper()
        return unit



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.DataModelBaseClasses.NetworkObject import NetworkObject
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Units.AttributeMappings.NetworkUnitMapping import NetworkUnits


@dataclass(repr=False)
class Drill(NetworkObject, ABC):

    @property
    @abstractmethod
    def total_drill_time(self) -> float:
        """The total time to drill the well."""
        raise NotImplementedError("Implement this in the derived class")

    def __init__(self, unit_system: Optional[UnitSystem] = None, name: Optional[str] = None, date: Optional[str] = None,
                 date_format: Optional[DateFormat] = None, start_date: Optional[str] = None,
                 properties_dict: Optional[DataObjectMixinDictType] = None) -> None:
        """Initialises the Drill class.

        Args:
            date (Optional[str]): The date of the object.
            date_format (Optional[DateFormat]): The date format that the object uses.
            start_date (Optional[str]): The start date of the model. Required if the object uses a decimal TIME.
            unit_system (Optional[UnitSystem]): The unit system of the object e.g. ENGLISH, METRIC.
            name (Optional[str]): The name of the object.
            properties_dict (dict): dict of the properties to set on the object.
        """

        super().__init__(unit_system=unit_system, name=name, date=date, date_format=date_format, start_date=start_date,
                         properties_dict=properties_dict)

    @property
    def units(self) -> NetworkUnits:
        """Returns the attribute to unit map for the Drill."""
        return NetworkUnits(self.unit_system)



from dataclasses import dataclass

import pandas as pd


@dataclass
class GridToProc:
    """Class for storing the GRIDTOPROC table information from the Options file."""
    grid_to_proc_table: None | pd.DataFrame = None
    auto_distribute: None | str = None

    @property
    def table_header(self) -> str:
        """Start of the GRIDTOPROC definition table."""
        return 'GRIDTOPROC'

    @property
    def table_footer(self) -> str:
        """End of the GRIDTOPROC definition table."""
        return 'END' + self.table_header

    def get_number_of_processors(self) -> int:
        """Returns the number of processors to use for the simulation.

        Returns:
        -------
            int: number of processors to use for the simulation
        """
        if self.grid_to_proc_table is None:
            return 0
        if 'PROCESS' in self.grid_to_proc_table.columns:
            return self.grid_to_proc_table['PROCESS'].max()
        else:
            return 0

    def to_string(self) -> str:
        """Returns the GRIDTOPROC table as a string."""
        if self.grid_to_proc_table is None and self.auto_distribute is None:
            return ''
        output_str = self.table_header + '\n'
        if self.auto_distribute is not None:
            output_str += f"AUTO {self.auto_distribute}\n"
        elif self.grid_to_proc_table is not None and not self.grid_to_proc_table.empty:
            output_str += self.grid_to_proc_table.to_string(index=False, header=True, justify='left') + '\n'
        output_str += self.table_footer + '\n'

        return output_str
